@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -41,10 +41,10 @@ except ImportError:
     vc_version = 0
 
 # The tuple giving the version number.
-version_tuple = (6, 99, 12, 2, vc_version)
+version_tuple = (6, 99, 14, 1, vc_version)
 
 # The name of this version.
-version_name = "We get the job done."
+version_name = "A funny thing happened."
 
 # A string giving the version number only (7.0.1.123).
 version_only = ".".join(str(i) for i in version_tuple)
@@ -71,6 +71,9 @@ linux = False
 android = False
 ios = False
 
+# Should we enable experimental features and debugging?
+experimental = "RENPY_EXPERIMENTAL" in os.environ
+
 import platform
 
 
@@ -151,6 +154,7 @@ backup_blacklist = {
     "renpy.object",
     "renpy.log",
     "renpy.bootstrap",
+    "renpy.debug",
     "renpy.display",
     "renpy.display.pgrender",
     "renpy.display.scale",
@@ -186,6 +190,8 @@ name_blacklist = {
     "renpy.display.render.IDENTITY",
     "renpy.loader.auto_lock",
     "renpy.display.screen.cprof",
+    "renpy.audio.audio.lock",
+    "renpy.audio.audio.periodic_condition",
     }
 
 
@@ -291,6 +297,7 @@ class Backup():
             mod, field = k
             setattr(mod, field, objects[v])
 
+
 # A backup of the Ren'Py modules after initial import.
 backup = None
 
@@ -316,22 +323,33 @@ def update_path(package):
     libexec = os.path.dirname(encodings.__path__[0])
     package.__path__.append(os.path.join(libexec, *name))
 
+# Replaced below.
+
+
+def plog(level, even, *args):
+    return
+
 
 def import_all():
 
     # Note: If we add a new update_path, we have to add an equivalent
     # hook in the renpython hooks dir.
 
+    # Note: If we add a new module, we need to add it to iOS.
+
     import renpy  # @UnresolvedImport
 
     update_path(renpy)
 
     import renpy.arguments  # @UnresolvedImport
 
+    import renpy.config
     import renpy.log
 
     import renpy.display
 
+    import renpy.debug
+
     # Should probably be early, as we will add it as a base to serialized things.
     import renpy.object
 
@@ -354,10 +372,15 @@ def import_all():
     import renpy.persistent
     import renpy.scriptedit
     import renpy.parser
+    import renpy.performance
+    import renpy.pydict
     import renpy.python
     import renpy.script
     import renpy.statements
 
+    global plog
+    plog = renpy.performance.log
+
     import renpy.styledata  # @UnresolvedImport
     update_path(renpy.styledata)
 
@@ -465,7 +488,6 @@ def import_all():
     import renpy.add_from
     import renpy.dump
 
-    import renpy.config  # depends on lots. @UnresolvedImport
     import renpy.minstore  # depends on lots. @UnresolvedImport
     import renpy.defaultstore  # depends on everything. @UnresolvedImport
 
@@ -554,6 +576,8 @@ def reload_all():
     renpy.display.draw = None
     renpy.display.interface = None
 
+    py_compile_cache = renpy.python.py_compile_cache
+
     # Delete the store modules.
     for i in sys.modules.keys():
         if i.startswith("store") or i == "renpy.store":
@@ -567,6 +591,8 @@ def reload_all():
     # Restore the state of all modules from backup.
     backup.restore()
 
+    renpy.python.old_py_compile_cache = py_compile_cache
+
     renpy.display.im.reset_module()
 
     post_import()
