@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,10 +24,15 @@
 import renpy.display
 
 import re
+import os
+import collections
 
 # This matches the dialogue-relevant text tags.
 TAG_RE = re.compile(r'(\{\{)|(\{(p|w|nw|fast)(?:\=([^}]*))?\})', re.S)
 
+less_pauses = ("RENPY_LESS_PAUSES" in os.environ)
+
+
 class DialogueTextTags(object):
     """
     This object parses the text tags that only make sense in dialogue,
@@ -71,9 +76,10 @@ class DialogueTextTags(object):
                     continue
 
                 if tag == "p" or tag == "w":
-                    self.pause_start.append(len(self.text))
-                    self.pause_end.append(len(self.text))
-                    self.pause_delay.append(value)
+                    if not less_pauses:
+                        self.pause_start.append(len(self.text))
+                        self.pause_end.append(len(self.text))
+                        self.pause_delay.append(value)
 
                 elif tag == "nw":
                     self.no_wait = True
@@ -97,7 +103,7 @@ class DialogueTextTags(object):
             self.pause_delay.append(None)
 
 
-def predict_show_display_say(who, what, who_args, what_args, window_args, image=False, two_window=False, side_image=None, screen=None, **kwargs):
+def predict_show_display_say(who, what, who_args, what_args, window_args, image=False, two_window=False, side_image=None, screen=None, properties=None, **kwargs):
     """
     This is the default function used by Character to predict images that
     will be used by show_display_say. It's called with more-or-less the
@@ -118,7 +124,7 @@ def predict_show_display_say(who, what, who_args, what_args, window_args, image=
         kwargs["image"] = image
 
     if screen:
-        props = compute_widget_properties(who_args, what_args, window_args)
+        props = compute_widget_properties(who_args, what_args, window_args, properties)
 
         renpy.display.predict.screen(
             screen,
@@ -132,27 +138,33 @@ def predict_show_display_say(who, what, who_args, what_args, window_args, image=
         return
 
 
-
-def compute_widget_properties(who_args, what_args, window_args, variant=None):
+def compute_widget_properties(who_args, what_args, window_args, properties, variant=None, multiple=None):
     """
     Computes and returns the widget properties.
     """
 
-    def style_args(d):
+    def style_args(d, name):
 
-        if not "style" in d:
-            return d
+        style = d.get("style", None)
+
+        if style is None:
+            if multiple is None:
+                return d
+            else:
+                style = name
 
         in_rollback = renpy.exports.in_rollback()
 
-        if (not in_rollback) and (not variant):
+        if (not in_rollback) and (not variant) and (not multiple):
             return d
 
         d = d.copy()
 
-        style = d["style"]
-
         if isinstance(style, basestring):
+
+            if multiple is not None:
+                style = "block{}_multiple{}_{}".format(multiple[0], multiple[1], style)
+
             style = getattr(renpy.store.style, style)
 
             if variant is not None:
@@ -165,15 +177,20 @@ def compute_widget_properties(who_args, what_args, window_args, variant=None):
 
         return d
 
-    who_args = style_args(who_args)
-    what_args = style_args(what_args)
-    window_args = style_args(window_args)
+    who_args = style_args(who_args, "who")
+    what_args = style_args(what_args, "what")
+    window_args = style_args(window_args, "window")
 
-    return {
-        "window" : window_args,
-        "what" : what_args,
-        "who" : who_args,
-        }
+    rv = dict(properties)
+
+    for prefix in renpy.config.character_id_prefixes:
+        rv[prefix] = style_args(properties.get(prefix, {}), prefix)
+
+    rv["window"] = window_args
+    rv["what"] = what_args
+    rv["who"] = who_args
+
+    return rv
 
 
 def show_display_say(who, what, who_args={}, what_args={}, window_args={},
@@ -185,6 +202,8 @@ def show_display_say(who, what, who_args={}, what_args={}, window_args={},
                      variant=None,
                      screen=None,
                      layer=None,
+                     properties={},
+                     multiple=None,
                      **kwargs):
     """
     This is called (by default) by renpy.display_say to add the
@@ -219,7 +238,7 @@ def show_display_say(who, what, who_args={}, what_args={}, window_args={},
     displaying the what text.
     """
 
-    props = compute_widget_properties(who_args, what_args, window_args, variant=variant)
+    props = compute_widget_properties(who_args, what_args, window_args, properties, variant=variant, multiple=multiple)
 
     def handle_who():
         if who:
@@ -251,20 +270,28 @@ def show_display_say(who, what, who_args={}, what_args={}, window_args={},
         tag = screen
         index = 0
 
-        while renpy.exports.showing(tag):
-            index += 1
-            tag = "%s%d" % (screen, index)
+        if multiple:
+
+            if renpy.display.screen.has_screen("multiple_" + screen):
+                screen = "multiple_" + screen
+                kwargs["multiple"] = multiple
+
+            tag = "block{}_multiple{}_{}".format(multiple[0], multiple[1], tag)
 
         if image:
             kwargs["image"] = image
 
+        if (side_image is not None) or renpy.config.old_say_args:
+            kwargs["side_image"] = side_image
+
+        if two_window or renpy.config.old_say_args:
+            kwargs["two_window"] = two_window
+
         renpy.display.screen.show_screen(
             screen,
             _widget_properties=props,
-            _transient = True,
-            _tag = tag,
-            side_image=side_image,
-            two_window=two_window,
+            _transient=True,
+            _tag=tag,
             who=who,
             what=what,
             _layer=layer,
@@ -272,8 +299,7 @@ def show_display_say(who, what, who_args={}, what_args={}, window_args={},
 
         renpy.exports.shown_window()
 
-        return renpy.display.screen.get_widget(screen, "what", layer)
-
+        return (tag, "what", layer)
 
     # Apply the transform.
     if transform:
@@ -311,11 +337,10 @@ def show_display_say(who, what, who_args={}, what_args={}, window_args={},
     return rv
 
 
-
 class SlowDone(object):
     delay = None
 
-    def __init__(self, ctc, ctc_position, callback, interact, type, cb_args, delay): #@ReservedAssignment
+    def __init__(self, ctc, ctc_position, callback, interact, type, cb_args, delay):  # @ReservedAssignment
         self.ctc = ctc
         self.ctc_position = ctc_position
         self.callback = callback
@@ -326,12 +351,22 @@ class SlowDone(object):
 
     def __call__(self):
 
-        if self.ctc and self.ctc_position == "fixed":
+        if renpy.display.screen.has_screen("ctc"):
+
+            if self.ctc:
+                args = [ self.ctc ]
+            else:
+                args = [ ]
+
+            renpy.display.screen.show_screen("ctc", *args, _transient=True)
+            renpy.exports.restart_interaction()
+
+        elif self.ctc and self.ctc_position == "fixed":
             renpy.display.screen.show_screen("_ctc", _transient=True, ctc=self.ctc)
             renpy.exports.restart_interaction()
 
         if self.delay is not None:
-            renpy.ui.pausebehavior(self.delay, True)
+            renpy.ui.pausebehavior(self.delay, True, voice=True)
             renpy.exports.restart_interaction()
 
         for c in self.callback:
@@ -340,31 +375,53 @@ class SlowDone(object):
 
 # This function takes care of repeatably showing the screen as part of
 # an interaction.
-def display_say(
-    who,
-    what,
-    show_function,
-    interact,
-    slow,
-    afm,
-    ctc,
-    ctc_pause,
-    ctc_position,
-    all_at_once,
-    cb_args,
-    with_none,
-    callback,
-    type, #@ReservedAssignment
-    checkpoint=True,
-    ctc_timedpause=None,
-    ctc_force=False):
 
-    if interact and (not renpy.game.preferences.skip_unseen) and (not renpy.game.context().seen_current(True)) and renpy.config.skipping == "fast":
+
+def display_say(
+        who,
+        what,
+        show_function,
+        interact,
+        slow,
+        afm,
+        ctc,
+        ctc_pause,
+        ctc_position,
+        all_at_once,
+        cb_args,
+        with_none,
+        callback,
+        type,  # @ReservedAssignment
+        checkpoint=True,
+        ctc_timedpause=None,
+        ctc_force=False,
+        advance=True,
+        multiple=None):
+
+    # Final is true if this statement should perform an interaction.
+
+    if multiple is None:
+        final = interact
+    else:
+        step, total = multiple
+
+        if step == total:
+            final = interact
+        else:
+            final = False
+
+    if not final:
+        advance = False
+
+    if final and (not renpy.game.preferences.skip_unseen) and (not renpy.game.context().seen_current(True)) and renpy.config.skipping == "fast":
         renpy.config.skipping = None
 
     # If we're in fast skipping mode, don't bother with say
     # statements at all.
-    if interact and renpy.config.skipping == "fast":
+    if advance and renpy.config.skipping == "fast":
+
+        for i in renpy.config.fast_skipping_callbacks:
+            i()
 
         # Clears out transients.
         renpy.exports.with_statement(None)
@@ -386,10 +443,10 @@ def display_say(
     for c in callback:
         c("begin", interact=interact, type=type, **cb_args)
 
-    if renpy.exports.roll_forward_info():
+    roll_forward = renpy.exports.roll_forward_info()
+
+    if roll_forward is True:
         roll_forward = False
-    else:
-        roll_forward = None
 
     # If we're just after a rollback or roll_forward, disable slow.
     after_rollback = renpy.game.after_rollback
@@ -399,6 +456,7 @@ def display_say(
 
     # If we're committed to skipping this statement, disable slow.
     elif (renpy.config.skipping and
+          advance and
           (renpy.game.preferences.skip_unseen or
            renpy.game.context().seen_current(True))):
         slow = False
@@ -420,109 +478,162 @@ def display_say(
         pause_end = dtt.pause_end
         pause_delay = dtt.pause_delay
 
+    exception = None
 
-    for i, (start, end, delay) in enumerate(zip(pause_start, pause_end, pause_delay)):
+    try:
 
-        last_pause = (i == len(pause_start) - 1)
+        for i, (start, end, delay) in enumerate(zip(pause_start, pause_end, pause_delay)):
 
-        # If we're going to do an interaction, then saybehavior needs
-        # to be here.
-        if interact:
-            behavior = renpy.ui.saybehavior(allow_dismiss=renpy.config.say_allow_dismiss)
-        else:
-            behavior = None
+            last_pause = (i == len(pause_start) - 1)
 
-
-        # The string to show.
-        what_string = dtt.text
-
-        # Figure out the CTC to use, if any.
-        if last_pause:
-            what_ctc = ctc
-        else:
-            if delay is not None:
-                what_ctc = ctc_timedpause or ctc_pause
+            # If we're going to do an interaction, then saybehavior needs
+            # to be here.
+            if advance:
+                behavior = renpy.ui.saybehavior(allow_dismiss=renpy.config.say_allow_dismiss)
             else:
-                what_ctc = ctc_pause
+                behavior = None
 
-        if not (interact or ctc_force):
-            what_ctc = None
+            # The string to show.
+            what_string = dtt.text
 
-        what_ctc = renpy.easy.displayable_or_none(what_ctc)
+            # Figure out the CTC to use, if any.
+            if last_pause:
+                what_ctc = ctc
+            else:
+                if delay is not None:
+                    what_ctc = ctc_timedpause or ctc_pause
+                else:
+                    what_ctc = ctc_pause
 
-        if what_ctc is not None:
-            what_ctc = what_ctc.parameterize(('ctc',), ())
+            if not (interact or ctc_force):
+                what_ctc = None
 
-        if delay == 0:
-            what_ctc = None
+            what_ctc = renpy.easy.displayable_or_none(what_ctc)
 
-        # Create the callback that is called when the slow text is done.
-        slow_done = SlowDone(what_ctc, ctc_position, callback, interact, type, cb_args, delay)
+            if (what_ctc is not None) and what_ctc._duplicatable:
+                what_ctc = what_ctc._duplicate(None)
+                what_ctc._unique()
 
-        # Run the show callback.
-        for c in callback:
-            c("show", interact=interact, type=type, **cb_args)
+            if delay == 0:
+                what_ctc = None
 
-        # Show the text.
-        what_text = show_function(who, what_string)
+            # Run the show callback.
+            for c in callback:
+                c("show", interact=interact, type=type, **cb_args)
 
-        if not isinstance(what_text, renpy.text.text.Text): #@UndefinedVariable
-            raise Exception("The say screen (or show_function) must return a Text object.")
+            # Create the callback that is called when the slow text is done.
+            slow_done = SlowDone(what_ctc, ctc_position, callback, interact, type, cb_args, delay)
 
-        if what_ctc and ctc_position == "nestled":
-            what_text.set_ctc(what_ctc)
+            # Show the text.
+            if multiple:
+                what_text = show_function(who, what_string, multiple=multiple)
+            else:
+                what_text = show_function(who, what_string)
 
-        # Update the properties of the what_text widget.
-        what_text.start = start
-        what_text.end = end
-        what_text.slow = slow
-        what_text.slow_done = slow_done
+            if interact or what_string or (what_ctc is not None) or (behavior and afm):
 
-        what_text.update()
+                if isinstance(what_text, tuple):
+                    what_text = renpy.display.screen.get_widget(what_text[0], what_text[1], what_text[2])
 
-        for c in callback:
-            c("show_done", interact=interact, type=type, **cb_args)
+                if not isinstance(what_text, renpy.text.text.Text):  # @UndefinedVariable
+                    raise Exception("The say screen (or show_function) must return a Text object.")
 
-        if behavior and afm:
-            behavior.set_text(what_text)
+                if what_ctc and ctc_position == "nestled":
+                    what_text.set_ctc(what_ctc)
 
-        if not slow:
-            slow_done()
+                # Update the properties of the what_text widget.
+                what_text.start = start
+                what_text.end = end
+                what_text.slow = slow
+                what_text.slow_done = slow_done
 
-        if interact:
-            rv = renpy.ui.interact(mouse='say', type=type, roll_forward=roll_forward)
+                what_text.update()
 
-            # This is only the case if the user has rolled forward, {nw} happens, or
-            # maybe in some other obscure cases.
-            if rv is False:
-                break
+                if behavior and afm:
+                    behavior.set_text(what_text)
 
-            if not last_pause:
-                for i in renpy.config.say_sustain_callbacks:
-                    i()
+            else:
+
+                slow = False
+
+            for c in callback:
+                c("show_done", interact=interact, type=type, **cb_args)
+
+            if not slow:
+                slow_done()
+
+            if final:
+                rv = renpy.ui.interact(mouse='say', type=type, roll_forward=roll_forward)
+
+                # This is only the case if the user has rolled forward, {nw} happens, or
+                # maybe in some other obscure cases.
+                if rv is False:
+                    break
+
+                if isinstance(rv, (renpy.game.JumpException, renpy.game.CallException)):
+                    raise rv
+
+                if not last_pause:
+                    for i in renpy.config.say_sustain_callbacks:
+                        i()
+
+    except (renpy.game.JumpException, renpy.game.CallException) as e:
+
+        exception = e
 
     # Do the checkpoint and with None.
-    if interact:
+    if final:
 
         if not dtt.no_wait:
             if checkpoint:
-                renpy.exports.checkpoint(True)
+                if exception is None:
+                    renpy.exports.checkpoint(True)
+                else:
+                    renpy.exports.checkpoint(exception)
+
         else:
             renpy.game.after_rollback = after_rollback
 
         if with_none is None:
             with_none = renpy.config.implicit_with_none
 
+        renpy.plog(1, "before with none")
+
         if with_none:
             renpy.game.interface.do_with(None, None)
 
+        renpy.plog(1, "after with none")
+
     for c in callback:
         c("end", interact=interact, type=type, **cb_args)
 
+    if exception is not None:
+        raise
+
+
+class HistoryEntry(renpy.object.Object):
+    """
+    Instances of this object are used to represent history entries in
+    _history_list.
+    """
+
+    # See ADVCharacter.add_history for the fields.
+
+    multiple = None
+
+    def __repr__(self):
+        return "<History {!r} {!r}>".format(self.who, self.what)
+
 
 # This is used to flag values that haven't been set by the user.
 NotSet = renpy.object.Sentinel("NotSet")
 
+
+# The number of multiple characters we've seen during the current
+# interaction.
+multiple_count = 0
+
+
 class ADVCharacter(object):
     """
     The character object contains information about a character. When
@@ -543,13 +654,14 @@ class ADVCharacter(object):
         ]
 
     voice_tag = None
+    properties = { }
 
     # When adding a new argument here, remember to add it to copy below.
     def __init__(
-        self,
-        name=NotSet,
-        kind=None,
-        **properties):
+            self,
+            name=NotSet,
+            kind=None,
+            **properties):
 
         if kind is None:
             kind = renpy.store.adv
@@ -565,7 +677,6 @@ class ADVCharacter(object):
             else:
                 return getattr(kind, n)
 
-
         # Similar, but it grabs the value out of kind.display_args instead.
         def d(n):
             if n in properties:
@@ -598,19 +709,22 @@ class ADVCharacter(object):
             self.image_tag = None
 
         self.display_args = dict(
-            interact = d('interact'),
-            slow = d('slow'),
-            afm = d('afm'),
-            ctc = renpy.easy.displayable_or_none(d('ctc')),
-            ctc_pause = renpy.easy.displayable_or_none(d('ctc_pause')),
-            ctc_timedpause = renpy.easy.displayable_or_none(d('ctc_timedpause')),
-            ctc_position = d('ctc_position'),
-            all_at_once = d('all_at_once'),
-            with_none = d('with_none'),
-            callback = d('callback'),
-            type = d('type'),
+            interact=d('interact'),
+            slow=d('slow'),
+            afm=d('afm'),
+            ctc=renpy.easy.displayable_or_none(d('ctc')),
+            ctc_pause=renpy.easy.displayable_or_none(d('ctc_pause')),
+            ctc_timedpause=renpy.easy.displayable_or_none(d('ctc_timedpause')),
+            ctc_position=d('ctc_position'),
+            all_at_once=d('all_at_once'),
+            with_none=d('with_none'),
+            callback=d('callback'),
+            type=d('type'),
+            advance=d('advance'),
             )
 
+        self.properties = collections.defaultdict(dict)
+
         if kind:
             self.who_args = kind.who_args.copy()
             self.what_args = kind.what_args.copy()
@@ -618,6 +732,9 @@ class ADVCharacter(object):
             self.show_args = kind.show_args.copy()
             self.cb_args = kind.cb_args.copy()
 
+            for k, v in kind.properties.items():
+                self.properties[k] = dict(v)
+
         else:
             self.who_args = { "substitute" : False }
             self.what_args = { "substitute" : False }
@@ -632,55 +749,65 @@ class ADVCharacter(object):
         if "slow_abortable" in properties:
             self.what_args["slow_abortable"] = properties.pop("slow_abortable")
 
-        for k in list(properties):
+        prefixes = [ "show", "cb", "what", "window", "who"] + renpy.config.character_id_prefixes
+        split_args = [ i + "_" for i in prefixes ] + [ "" ]
 
-            if "_" in k:
-                prefix, suffix = k.split("_", 1)
+        split = renpy.easy.split_properties(properties, *split_args)
 
-                if prefix == "show":
-                    self.show_args[suffix] = properties[k]
-                    continue
-                elif prefix == "cb":
-                    self.cb_args[suffix] = properties[k]
-                    continue
-                elif prefix == "what":
-                    self.what_args[suffix] = properties[k]
-                    continue
-                elif prefix == "window":
-                    self.window_args[suffix] = properties[k]
-                    continue
-                elif prefix == "who":
-                    self.who_args[suffix] = properties[k]
-                    continue
+        for prefix, d in zip(prefixes, split):
+            self.properties[prefix].update(d)
 
-            self.who_args[k] = properties[k]
+        self.properties["who"].update(split[-1])
+
+        self.show_args.update(self.properties.pop("show"))
+        self.cb_args.update(self.properties.pop("cb"))
+        self.what_args.update(self.properties.pop("what"))
+        self.window_args.update(self.properties.pop("window"))
+        self.who_args.update(self.properties.pop("who"))
 
     def copy(self, name=NotSet, **properties):
         return type(self)(name, kind=self, **properties)
 
     # This is called before the interaction.
-    def do_add(self, who, what):
+    def do_add(self, who, what, multiple=None):
         return
 
     # This is what shows the screen for a given interaction.
-    def do_show(self, who, what):
-        return self.show_function(
-            who,
-            what,
-            who_args=self.who_args,
-            what_args=self.what_args,
-            window_args=self.window_args,
-            screen=self.screen,
-            **self.show_args)
+    def do_show(self, who, what, multiple=None):
+
+        if multiple is not None:
+
+            return self.show_function(
+                who,
+                what,
+                who_args=self.who_args,
+                what_args=self.what_args,
+                window_args=self.window_args,
+                screen=self.screen,
+                properties=self.properties,
+                multiple=multiple,
+                **self.show_args)
+
+        else:
+
+            return self.show_function(
+                who,
+                what,
+                who_args=self.who_args,
+                what_args=self.what_args,
+                window_args=self.window_args,
+                screen=self.screen,
+                properties=self.properties,
+                **self.show_args)
 
     # This is called after the last interaction is done.
-    def do_done(self, who, what):
-        return
+    def do_done(self, who, what, multiple=None):
+        self.add_history("adv", who, what, multiple=multiple)
 
     # This is called when an extend occurs, before the usual add/show
     # cycel.
     def do_extend(self):
-        return
+        self.pop_history()
 
     # This is called to actually do the displaying.
     def do_display(self, who, what, **display_args):
@@ -689,7 +816,6 @@ class ADVCharacter(object):
                     self.do_show,
                     **display_args)
 
-
     # This is called to predict images that will be used by this
     # statement.
     def do_predict(self, who, what):
@@ -700,6 +826,7 @@ class ADVCharacter(object):
             what_args=self.what_args,
             window_args=self.window_args,
             screen=self.screen,
+            properties=self.properties,
             **self.show_args)
 
     def resolve_say_attributes(self, predict, wanted=[], remove=[]):
@@ -742,31 +869,53 @@ class ADVCharacter(object):
                 images.predict_show(new_image)
             else:
                 trans = renpy.config.say_attribute_transition
+                layer = renpy.config.say_attribute_transition_layer
 
-                if trans is not None:
+                if (trans is not None) and (layer is not None):
                     renpy.exports.with_statement(None)
 
                 renpy.exports.show(new_image)
 
                 if trans is not None:
-                    renpy.exports.with_statement(trans)
+                    if layer is None:
+                        renpy.exports.with_statement(trans)
+                    else:
+                        renpy.exports.transition(trans, layer=layer)
 
         else:
 
             # Otherwise, just record the attributes of the image.
             images.predict_show("master", tagged_attrs, show=False)
 
-    def __str__(self):
+    def __unicode__(self):
+
         who = self.name
 
-        # If dynamic is set, evaluate the name expression.
         if self.dynamic:
             who = renpy.python.py_eval(who)
 
-        return who
+        return renpy.substitutions.substitute(who)[0]
 
+    def __str__(self):
+        return unicode(self).encode("utf-8")
 
-    def __call__(self, what, interact=True, **kwargs):
+    def __format__(self, spec):
+        return format(unicode(self), spec)
+
+    def __repr__(self):
+        return "<Character: {!r}>".format(self.name)
+
+    def empty_window(self):
+        if renpy.config.fast_empty_window and (self.name is None) and not (self.what_prefix or self.what_suffix):
+            self.do_show(None, "")
+            return
+
+        self("", interact=False, _call_done=False)
+
+    def __call__(self, what, interact=True, _call_done=True, multiple=None, **kwargs):
+
+        if kwargs:
+            return Character(kind=self, **kwargs)(what, interact=interact, _call_done=_call_done, multiple=multiple)
 
         # Check self.condition to see if we should show this line at all.
         if not (self.condition is None or renpy.python.py_eval(self.condition)):
@@ -775,21 +924,38 @@ class ADVCharacter(object):
         if not isinstance(what, basestring):
             raise Exception("Character expects its what argument to be a string, got %r." % (what,))
 
-        if interact and (renpy.config.speaking_attribute is not None):
-            speaking = [ renpy.config.speaking_attribute ]
+        # Figure out multiple and final. Multiple is None if this is not a multiple
+        # dialogue, or a step and the total number of steps in a multiple interaction.
+
+        global multiple_count
+
+        if multiple is None:
+            multiple_count = 0
+
         else:
-            speaking = [ ]
+            multiple_count += 1
+            multiple = (multiple_count, multiple)
 
-        self.resolve_say_attributes(False, wanted=speaking)
+            if multiple_count == multiple[1]:
+                multiple_count = 0
 
-        old_side_image_attributes = renpy.store._side_image_attributes
+        if multiple is None:
 
-        if self.image_tag:
-            attrs = (self.image_tag,) + renpy.game.context().images.get_attributes("master", self.image_tag)
-        else:
-            attrs = None
+            if interact and (renpy.config.speaking_attribute is not None):
+                speaking = [ renpy.config.speaking_attribute ]
+            else:
+                speaking = [ ]
 
-        renpy.store._side_image_attributes = attrs
+            self.resolve_say_attributes(False, wanted=speaking)
+
+            old_side_image_attributes = renpy.store._side_image_attributes
+
+            if self.image_tag:
+                attrs = (self.image_tag,) + renpy.game.context().images.get_attributes("master", self.image_tag)
+            else:
+                attrs = None
+
+            renpy.store._side_image_attributes = attrs
 
         if renpy.config.voice_tag_callback is not None:
             renpy.config.voice_tag_callback(self.voice_tag)
@@ -801,9 +967,11 @@ class ADVCharacter(object):
 
             # Figure out the arguments to display.
             display_args = self.display_args.copy()
-            display_args.update(kwargs)
             display_args["interact"] = display_args["interact"] and interact
 
+            if multiple is not None:
+                display_args["multiple"] = multiple
+
             who = self.name
 
             # If dynamic is set, evaluate the name expression.
@@ -835,28 +1003,37 @@ class ADVCharacter(object):
 
             # Run the add_function, to add this character to the
             # things like NVL-mode.
-            self.do_add(who, what)
+
+            if multiple is not None:
+                self.do_add(who, what, multiple=multiple)
+            else:
+                self.do_add(who, what)
 
             # Now, display the damned thing.
             self.do_display(who, what, cb_args=self.cb_args, **display_args)
 
             # Indicate that we're done.
-            self.do_done(who, what)
+            if _call_done:
 
-            # Finally, log this line of dialogue.
-            if who and isinstance(who, (str, unicode)):
-                renpy.exports.log(who)
-            renpy.exports.log(what)
-            renpy.exports.log("")
+                if multiple is not None:
+                    self.do_done(who, what, multiple=multiple)
+                else:
+                    self.do_done(who, what)
+
+                # Finally, log this line of dialogue.
+                if who and isinstance(who, (str, unicode)):
+                    renpy.exports.log(who)
+
+                renpy.exports.log(what)
+                renpy.exports.log("")
 
         finally:
 
-            if interact:
+            if (multiple is None) and interact:
                 renpy.store._side_image_attributes = old_side_image_attributes
 
                 self.resolve_say_attributes(False, remove=speaking)
 
-
     def predict(self, what):
 
         self.resolve_say_attributes(True)
@@ -892,6 +1069,70 @@ class ADVCharacter(object):
 
         return self.display_args['interact']
 
+    def add_history(self, kind, who, what, multiple=None, **kwargs):
+        """
+        This is intended to be called by subclasses of ADVCharacter to add
+        History entries to _history_list.
+        """
+
+        history_length = renpy.config.history_length
+
+        if history_length is None:
+            return
+
+        if not renpy.store._history:  # @UndefinedVariable
+            return
+
+        history = renpy.store._history_list  # @UndefinedVariable
+
+        h = HistoryEntry()
+
+        h.kind = kind
+
+        h.who = who
+        h.what = what
+
+        h.who_args = self.who_args
+        h.what_args = self.what_args
+        h.window_args = self.window_args
+        h.show_args = self.show_args
+
+        h.image_tag = self.image_tag
+
+        h.multiple = multiple
+
+        if renpy.game.context().rollback:
+            h.rollback_identifier = renpy.game.log.current.identifier
+        else:
+            h.rollback_identifier = None
+
+        for k, v in kwargs.items():
+            setattr(h, k, v)
+
+        for i in renpy.config.history_callbacks:
+            i(h)
+
+        history.append(h)
+
+        while len(history) > history_length:
+            history.pop(0)
+
+    def pop_history(self):
+        """
+        This is intended to be called by do_extend to remove entries from
+        _history_list.
+        """
+
+        history_length = renpy.config.history_length
+
+        if history_length is None:
+            return
+
+        if not renpy.store._history:  # @UndefinedVariable
+            return
+
+        renpy.store._history_list.pop()  # @UndefinedVariable
+
 
 def Character(name=NotSet, kind=None, **properties):
     """
@@ -976,6 +1217,13 @@ def Character(name=NotSet, kind=None, **properties):
         dialogue is shown. If false, an interaction will not occur,
         and additional elements can be added to the screen.
 
+    `advance`
+        If True, the default, the player can click to advance through
+        the statement, and other means of advancing (such as skip and
+        auto-forward mode) will also work. If False, the player will be
+        unable to move past the say statement unless an alternate means
+        (such as a jump hyperlink or screen) is provided.
+
     `mode`
         A string giving the mode to enter when this character
         speaks. See the section on :ref:`modes <modes>` for more details.
@@ -1020,22 +1268,11 @@ def Character(name=NotSet, kind=None, **properties):
 
     Keyword arguments beginning with ``show_`` have the prefix
     stripped off, and are passed to the screen as arguments. For
-    example, the value of ``show_side_image`` will become the
-    value of the ``side_image`` variable in the screen.
+    example, the value of ``show_myflag`` will become the value of
+    the ``myflag`` variable in the screen. (The myflag variable isn't
+    used by default, but can be used by a custom say screen.)
 
-    Some useful ``show_`` variables implemented by the default screens are:
-
-    `show_side_image`
-        When given a Displayable, shows that displayable when the
-        dialogue is shown. The position of that displayable is
-        controlled by its position properties. This is often used
-        to show an image of the speaking character to the side
-        of the dialogue.
-
-    `show_two_window`
-        If true, restructures the layout so that the name of the
-        character is placed in one window, and the dialogue text in a
-        second window.
+    One show variable is, for historical reasons, handled by Ren'Py itself:
 
     `show_layer`
         If given, this should be a string giving the name of the layer
@@ -1056,7 +1293,12 @@ def Character(name=NotSet, kind=None, **properties):
     The style applied to the character name, spoken text, and window
     can also be set this way, using the `who_style`, `what_style`, and
     `window_style` arguments, respectively.
-     """
+
+    Setting :var:`config.character_id_prefixes` makes it possible to style
+    other displayables as well. For example, when the default GUI is used,
+    styles prefixed with ``namebox_`` are used to style the name of the
+    speaking character.
+    """
 
     if kind is None:
         kind = renpy.store.adv
@@ -1066,4 +1308,3 @@ def Character(name=NotSet, kind=None, **properties):
 
 def DynamicCharacter(name_expr, **properties):
     return Character(name_expr, dynamic=True, **properties)
-
