@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -61,6 +61,10 @@ _widget_properties = { }
 # The text rectangle, or None to use the automatic code.
 _text_rect = None
 
+# Are we in various menus?
+_menu = False
+main_menu = False
+
 
 class _Config(object):
 
@@ -91,6 +95,12 @@ class _Config(object):
         if name == "script_version":
             renpy.store._set_script_version(value)  # E1101 @UndefinedVariable
 
+        if name == "developer":
+            if value == "auto":
+                renpy.config.original_developer = value
+                renpy.config.developer = renpy.config.default_developer
+                return
+
         cvars[name] = value
 
     def __delattr__(self, name):
@@ -99,6 +109,7 @@ class _Config(object):
         else:
             delattr(renpy.config, name)
 
+
 # The styles object.
 style = None
 
@@ -230,6 +241,7 @@ class _layout_class(__builtins__["object"]):
 
         return rv
 
+
 Fixed = _layout_class(renpy.display.layout.MultiBox, """
 :doc: disp_box
 :args: (*args, **properties)
@@ -255,6 +267,7 @@ A layout that lays out its members from top to bottom.
 
 Grid = _layout_class(renpy.display.layout.Grid, """
 :doc: disp_grid
+:args: (cols, rows, *args, **properties)
 
 Lays out displayables in a grid. The first two positional arguments
 are the number of columns and rows in the grid. This must be followed
@@ -355,6 +368,7 @@ adv = ADVCharacter(None,
                    with_none=None,
                    callback=None,
                    type='say',
+                   advance=True,
 
                    who_style='say_label',
                    what_style='say_dialogue',
@@ -378,13 +392,16 @@ def predict_say(who, what):
         pass
 
 
-def say(who, what, interact=True):
+def say(who, what, interact=True, *args, **kwargs):
     who = Character(who, kind=adv)
-    who(what, interact=interact)
+    who(what, interact=interact, *args, **kwargs)
 
-# Used by renpy.reshow_say.
+
+# Used by renpy.reshow_say and extend.
 _last_say_who = None
 _last_say_what = None
+_last_say_args = ()
+_last_say_kwargs = { }
 
 # Used to store the things pinned into the cache.
 _cache_pin_set = set()
@@ -409,6 +426,10 @@ _side_image_attributes = None
 # among other things.
 main_menu = False
 
+# The action that's used when the player clicks the ignore button on the
+# error handling screen.
+_ignore_action = None
+
 # Make these available to user code.
 import sys
 import os
@@ -423,4 +444,5 @@ def public_api():
     os
     sys
 
+
 del public_api
