@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -53,7 +53,7 @@ def default_drag_joined(drag):
 class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
     """
     :doc: drag_drop class
-    :args: (d=None, drag_name=None, draggable=True, droppable=True, drag_raise=True, dragged=None, dropped=None, drag_handle=(0.0, 0.0, 1.0, 1.0), drag_joined=..., clicked=None, hovered=None, unhovered=None, **properties)
+    :args: (d=None, drag_name=None, draggable=True, droppable=True, drag_raise=True, dragged=None, dropped=None, drag_handle=(0.0, 0.0, 1.0, 1.0), drag_joined=..., clicked=None, hovered=None, unhovered=None, mouse_drop=False, **properties)
 
     A displayable that represents an object that can be dragged around
     its enclosing area. A Drag can also represent an area that
@@ -149,6 +149,11 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         size, as the drags can leave the screen entirely, with no
         way to get them back on the screen.
 
+    `mouse_drop`
+        If true, the drag is dropped on the first droppable under the cursor. 
+        If false, the default, the drag is dropped onto the droppable with 
+        the largest degree of overlap.
+
     Except for `d`, all of the parameters are available as fields (with
     the same name) on the Drag object. In addition, after the drag has
     been rendered, the following fields become available:
@@ -181,6 +186,7 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
                  unhovered=None,
                  replaces=None,
                  drag_offscreen=False,
+                 mouse_drop=False,
                  style="drag",
                  **properties):
 
@@ -198,6 +204,10 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         self.hovered = hovered
         self.unhovered = unhovered
         self.drag_offscreen = drag_offscreen
+        # if mouse_drop_check is True (default False), the drop will not
+        #  use default major overlap between droppables but instead
+        #  will use mouse coordinates to select droppable
+        self.mouse_drop = mouse_drop
 
         # We're focusable if we can be dragged.
         self.focusable = draggable
@@ -272,6 +282,7 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
             self.old_position = replaces.old_position
             self.drag_moved = replaces.drag_moved
             self.last_drop = replaces.last_drop
+            self.mouse_drop = replaces.mouse_drop
 
         if d is not None:
             self.add(d)
@@ -345,6 +356,16 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
         if self.drag_group is not None:
             self.drag_group.raise_children([ self ])
 
+    def bottom(self):
+        """
+        :doc: drag_drop method
+
+        Lowers this displayable to the bottom of its drag_group.
+        """
+
+        if self.drag_group is not None:
+            self.drag_group.lower_children([ self ])
+
     def visit(self):
         return [ self.child ]
 
@@ -581,7 +602,10 @@ class Drag(renpy.display.core.Displayable, renpy.python.RevertableObject):
             handled = False
 
         if (self.drag_group is not None) and self.drag_moved:
-            drop = self.drag_group.get_best_drop(joined)
+            if self.mouse_drop:
+               drop = self.drag_group.get_drop_at(joined, par_x, par_y)
+            else:
+               drop = self.drag_group.get_best_drop(joined)
         else:
             drop = None
 
@@ -752,6 +776,20 @@ class DragGroup(renpy.display.layout.MultiBox):
         self.children = self._list_type(children)
         self.offsets = self._list_type(offsets)
 
+    def lower_children(self, l):
+        """
+        Lowers the children in `l` to the bottom of this drag group, with
+        the one at the bottom being the lowest.
+        """
+
+        self.children.reverse()
+        self.offsets.reverse()
+
+        self.raise_children(l)
+
+        self.children.reverse()
+        self.offsets.reverse()
+
     def get_best_drop(self, joined):
         """
         Returns the droppable that the members of joined overlap the most.
@@ -773,6 +811,9 @@ class DragGroup(renpy.display.layout.MultiBox):
                 if not c.droppable:
                     continue
 
+                if c.x is None:
+                    continue
+
                 r2 = (c.x, c.y, c.w, c.h)
 
                 overlap = rect_overlap_area(r1, r2)
@@ -786,6 +827,26 @@ class DragGroup(renpy.display.layout.MultiBox):
         else:
             return rv
 
+    def get_drop_at(self, joined, x, y):
+        """
+        Returns the droppable that is exactly at x, y.
+        """
+
+        joined_set = set(joined)
+        for c in self.children:
+            if c in joined_set:
+               continue
+
+            if not c.droppable:
+               continue
+
+            if c.x is None:
+               continue
+
+            if (x >= c.x and y >= c.y and
+                x < (c.x+c.w) and y < (c.y+c.h)):
+               return c
+
     def get_children(self):
         """
         Returns a list of Drags that are the children of
