@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,8 +23,7 @@
 
 import pygame_sdl2 as pygame
 import renpy.display
-import renpy.test
-import random
+
 
 class Focus(object):
 
@@ -80,7 +79,13 @@ focus_type = "mouse"
 # the focus to change.
 pending_focus_type = "mouse"
 
+
+# The current tooltip and tooltip screen.
+tooltip = None
+
 # Sets the currently focused widget.
+
+
 def set_focused(widget, arg, screen):
     global argument
     argument = arg
@@ -92,31 +97,77 @@ def set_focused(widget, arg, screen):
 
     renpy.display.tts.displayable(widget)
 
-# Gets the currently focused widget.
+    global tooltip
+
+    # Figure out the tooltip.
+
+    if widget is None:
+        new_tooltip = None
+    else:
+        new_tooltip = widget._tooltip
+
+    if tooltip != new_tooltip:
+        tooltip = new_tooltip
+        renpy.exports.restart_interaction()
+
+
 def get_focused():
+    """
+    Gets the currently focused displayable.
+    """
+
     return renpy.game.context().scene_lists.focused
 
-# Get the mouse cursor for the focused widget.
+
 def get_mouse():
+    """
+    Gets the mouse associated with the currently focused displayable.
+    """
+
     focused = get_focused()
     if focused is None:
         return None
     else:
         return focused.style.mouse
 
+
+def get_tooltip(screen=None):
+    """
+    Gets the tooltip information.
+    """
+
+    if screen is None:
+        return tooltip
+
+    if screen_of_focused is None:
+        return None
+
+    if screen_of_focused.screen_name[0] == screen:
+        return tooltip
+
+    if screen_of_focused.tag == screen:
+        return tooltip
+
+    return None
+
+
 def set_grab(widget):
     global grab
     grab = widget
 
     renpy.exports.cancel_gesture()
 
+
 def get_grab():
     return grab
 
+
 # The current list of focuses that we know about.
 focus_list = [ ]
 
 # This takes in a focus list from the rendering system.
+
+
 def take_focuses():
     global focus_list
     focus_list = [ ]
@@ -133,6 +184,7 @@ def take_focuses():
     if (default_focus is not None) and (get_focused() is None):
         change_focus(default_focus, True)
 
+
 def focus_coordinates():
     """
     :doc: other
@@ -154,13 +206,13 @@ def focus_coordinates():
 # A map from id(displayable) to the displayable that replaces it.
 replaced_by = { }
 
+
 def before_interact(roots):
     """
     Called before each interaction to choose the focused and grabbed
     displayables.
     """
 
-
     global new_grab
     global grab
 
@@ -253,6 +305,8 @@ def before_interact(roots):
 
 # This changes the focus to be the widget contained inside the new
 # focus object.
+
+
 def change_focus(newfocus, default=False):
     rv = None
 
@@ -296,6 +350,7 @@ def change_focus(newfocus, default=False):
 
     return rv
 
+
 def clear_focus():
     """
     Clears the focus when the window loses mouse focus.
@@ -304,6 +359,8 @@ def clear_focus():
     change_focus(None)
 
 # This handles mouse events, to see if they change the focus.
+
+
 def mouse_handler(ev, x, y, default=False):
     """
     Handle mouse events, to see if they change the focus.
@@ -320,7 +377,6 @@ def mouse_handler(ev, x, y, default=False):
         else:
             pending_focus_type = "mouse"
 
-
     new_focus = renpy.display.render.focus_at_point(x, y)
 
     if new_focus is None:
@@ -340,7 +396,7 @@ def focus_extreme(xmul, ymul, wmul, hmul):
 
     for f in focus_list:
 
-        if not f.x:
+        if f.x is None:
             continue
 
         score = (f.x * xmul +
