@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -103,6 +103,13 @@ class Container(renpy.display.core.Displayable):
 
         super(Container, self).__init__(**properties)
 
+    def _handles_event(self, event):
+        for i in self.children:
+            if i._handles_event(event):
+                return True
+
+        return False
+
     def set_style_prefix(self, prefix, root):
         super(Container, self).set_style_prefix(prefix, root)
 
@@ -120,6 +127,14 @@ class Container(renpy.display.core.Displayable):
         if rv.children:
             rv.child = rv.children[-1]
 
+        rv._duplicatable = False
+
+        for i in rv.children:
+            i._unique()
+
+            if i._duplicatable:
+                rv._duplicatable = True
+
         return rv
 
     def _in_current_store(self):
@@ -389,8 +404,15 @@ class Grid(Container):
 
     def render(self, width, height, st, at):
 
+        xspacing = self.style.xspacing
+        yspacing = self.style.yspacing
+
+        if xspacing is None:
+            xspacing = self.style.spacing
+        if yspacing is None:
+            yspacing = self.style.spacing
+
         # For convenience and speed.
-        padding = self.style.spacing
         cols = self.cols
         rows = self.rows
 
@@ -415,9 +437,9 @@ class Grid(Container):
         renheight = height
 
         if self.style.xfill:
-            renwidth = (width - (cols - 1) * padding) / cols
+            renwidth = (width - (cols - 1) * xspacing) / cols
         if self.style.yfill:
-            renheight = (height - (rows - 1) * padding) / rows
+            renheight = (height - (rows - 1) * yspacing) / rows
 
         renders = [ render(i, renwidth, renheight, st, at) for i in children ]
         sizes = [ i.get_size() for i in renders ]
@@ -435,8 +457,8 @@ class Grid(Container):
         if self.style.yfill:
             cheight = renheight
 
-        width = cwidth * cols + padding * (cols - 1)
-        height = cheight * rows + padding * (rows - 1)
+        width = cwidth * cols + xspacing * (cols - 1)
+        height = cheight * rows + yspacing * (rows - 1)
 
         rv = renpy.display.render.Render(width, height)
 
@@ -448,8 +470,8 @@ class Grid(Container):
                 child = children[ x + y * cols ]
                 surf = renders[x + y * cols]
 
-                xpos = x * (cwidth + padding)
-                ypos = y * (cheight + padding)
+                xpos = x * (cwidth + xspacing)
+                ypos = y * (cheight + yspacing)
 
                 offset = child.place(rv, xpos, ypos, cwidth, cheight, surf)
                 offsets.append(offset)
@@ -479,6 +501,7 @@ class MultiBox(Container):
     layer_name = None
     first = True
     order_reverse = False
+    layout = None
 
     def __init__(self, spacing=None, layout=None, style='default', **properties):
 
@@ -487,6 +510,8 @@ class MultiBox(Container):
 
         super(MultiBox, self).__init__(style=style, **properties)
 
+        self._clipping = self.style.clipping
+
         self.default_layout = layout
 
         # The start and animation times for children of this
@@ -599,6 +624,16 @@ class MultiBox(Container):
 
         self.scene_list.extend(l)
 
+    def update_times(self):
+
+        it = renpy.game.interface.interact_time
+
+        self.start_times = [ i or it for i in self.start_times ]
+        self.anim_times = [ i or it for i in self.anim_times ]
+
+        if it is None:
+            self.first = True
+
     def render(self, width, height, st, at):
 
         # Do we need to adjust the child times due to our being a layer?
@@ -620,28 +655,19 @@ class MultiBox(Container):
             self.first = False
 
             if adjust_times:
+                self.update_times()
 
-                it = renpy.game.interface.interact_time
+        layout = self.style.box_layout
 
-                self.start_times = [ i or it for i in self.start_times ]
-                self.anim_times = [ i or it for i in self.anim_times ]
-
-            layout = self.style.box_layout
-
-            if layout is None:
-                layout = self.default_layout
-
-            self.layout = layout  # W0201
-
-        else:
-            layout = self.layout
+        if layout is None:
+            layout = self.default_layout
 
         # Handle time adjustment, store the results in csts and cats.
         if adjust_times:
             t = renpy.game.interface.frame_time
 
-            csts = [ t - start for start in self.start_times ]
-            cats = [ t - anim for anim in self.anim_times ]
+            csts = [ 0 if (start is None) else (t - start) for start in self.start_times ]
+            cats = [ 0 if (anim is None) else (t - anim) for anim in self.anim_times ]
 
         else:
             csts = [ st ] * len(self.children)
@@ -935,6 +961,14 @@ class MultiBox(Container):
 
     def event(self, ev, x, y, st):
 
+        # Do we need to adjust the child times due to our being a layer?
+        if self.first:
+
+            self.first = False
+
+            if self.layer_name or (self.layers is not None):
+                self.update_times()
+
         children_offsets = zip(self.children, self.offsets, self.start_times)
 
         if not self.style.order_reverse:
@@ -1041,6 +1075,9 @@ class Window(Container):
         xminimum = scale(style.xminimum, width)
         yminimum = scale(style.yminimum, height)
 
+        xmaximum = scale(style.xmaximum, width)
+        ymaximum = scale(style.ymaximum, height)
+
         size_group = self.style.size_group
         if size_group and size_group in size_groups:
             xminimum = max(xminimum, size_groups[size_group].width(width, height, st, at))
@@ -1084,11 +1121,11 @@ class Window(Container):
 
         if renpy.config.enforce_window_max_size:
 
-            if style.xmaximum is not None:
-                width = min(width, style.xmaximum)
+            if xmaximum is not None:
+                width = min(width, xmaximum)
 
-            if style.ymaximum is not None:
-                height = min(height, style.ymaximum)
+            if ymaximum is not None:
+                height = min(height, ymaximum)
 
         rv = renpy.display.render.Render(width, height)
 
@@ -1160,18 +1197,18 @@ class DynamicDisplayable(renpy.display.core.Displayable):
 
     ::
 
-        # If tooltip is not empty, shows it in a text. Otherwise,
-        # show Null. Checks every tenth of a second to see if the
-        # tooltip has been updated.
+        # Shows a countdown from 5 to 0, updating it every tenth of
+        # a second until the time expires.
         init python:
-             def show_tooltip(st, at):
-                 if tooltip:
-                     return tooltip, .1
-                 else:
-                     return Null(), .1
 
-        image tooltipper = DynamicDisplayable(show_tooltip)
+            def show_countdown(st, at):
+                if st > 5.0:
+                    return Text("0.0"), None
+                else:
+                    d = Text("{:.1f}".format(5.0 - st))
+                    return d, 0.1
 
+        image countdown = DynamicDisplayable(show_countdown)
     """
 
     nosave = [ 'child' ]
@@ -1222,8 +1259,13 @@ class DynamicDisplayable(renpy.display.core.Displayable):
             else:
                 child, _ = self.function(0, 0, *self.args, **self.kwargs)
 
-            if child is not None:
-                renpy.display.predict.displayable(child)
+            if isinstance(child, list):
+
+                for i in child:
+                    renpy.display.predict.displayable(i)
+                else:
+                    renpy.display.predict.displayable(child)
+
         except:
             pass
 
@@ -1237,6 +1279,7 @@ class DynamicDisplayable(renpy.display.core.Displayable):
         if self.child:
             return self.child.event(ev, x, y, st)
 
+
 # A cache of compiled conditions used by ConditionSwitch.
 cond_cache = { }
 
@@ -1262,13 +1305,16 @@ def condition_switch_pick(switch):
     raise Exception("Switch could not choose a displayable.")
 
 
-def condition_switch_show(st, at, switch):
+def condition_switch_show(st, at, switch, predict_all=None):
     return condition_switch_pick(switch), None
 
 
-def condition_switch_predict(switch):
+def condition_switch_predict(switch, predict_all=None):
 
-    if renpy.game.lint:
+    if predict_all is None:
+        predict_all = renpy.config.conditionswitch_predict_all
+
+    if renpy.game.lint or (predict_all and renpy.display.predict.predicting):
         return [ d for _cond, d in switch ]
 
     return [ condition_switch_pick(switch) ]
@@ -1277,17 +1323,26 @@ def condition_switch_predict(switch):
 def ConditionSwitch(*args, **kwargs):
     """
     :doc: disp_dynamic
+    :args: (*args, predict_all=None, **properties)
 
     This is a displayable that changes what it is showing based on
-    python conditions. The positional argument should be given in
+    Python conditions. The positional arguments should be given in
     groups of two, where each group consists of:
 
-    * A string containing a python condition.
+    * A string containing a Python condition.
     * A displayable to use if the condition is true.
 
     The first true condition has its displayable shown, at least
     one condition should always be true.
 
+    The conditions uses here should not have externally-visible side-effects.
+
+    `predict_all`
+        If True, all of the possible displayables will be predicted when
+        the displayable is shown. If False, only the current condition is
+        predicted. If None, :var:`config.conditionswitch_predict_all` is
+        used.
+
     ::
 
         image jill = ConditionSwitch(
@@ -1295,6 +1350,7 @@ def ConditionSwitch(*args, **kwargs):
             "True", "jill_sober.png")
     """
 
+    predict_all = kwargs.pop("predict_all", None)
     kwargs.setdefault('style', 'default')
 
     switch = [ ]
@@ -1313,6 +1369,7 @@ def ConditionSwitch(*args, **kwargs):
 
     rv = DynamicDisplayable(condition_switch_show,
                             switch,
+                            predict_all,
                             _predict_function=condition_switch_predict)
 
     return Position(rv, **kwargs)
@@ -1321,6 +1378,7 @@ def ConditionSwitch(*args, **kwargs):
 def ShowingSwitch(*args, **kwargs):
     """
     :doc: disp_dynamic
+    :args: (*args, predict_all=None, **properties)
 
     This is a displayable that changes what it is showing based on the
     images are showing on the screen. The positional argument should
@@ -1331,6 +1389,12 @@ def ShowingSwitch(*args, **kwargs):
 
     A default image should be specified.
 
+    `predict_all`
+        If True, all of the possible displayables will be predicted when
+        the displayable is shown. If False, only the current condition is
+        predicted. If None, :var:`config.conditionswitch_predict_all` is
+        used.
+
     One use of ShowingSwitch is to have side images change depending on
     the current emotion of a character. For example::
 
@@ -1632,16 +1696,31 @@ class AdjustTimes(Container):
 
         self.add(child)
 
+    def adjusted_times(self):
+
+        interact_time = renpy.game.interface.interact_time
+
+        if (self.start_time is None) and (interact_time is not None):
+            self.start_time = interact_time
+
+        if self.start_time is not None:
+            st = renpy.game.interface.frame_time - self.start_time
+        else:
+            st = 0
+
+        if (self.anim_time is None) and (interact_time is not None):
+            self.anim_time = interact_time
+
+        if self.anim_time is not None:
+            at = renpy.game.interface.frame_time - self.anim_time
+        else:
+            at = 0
+
+        return st, at
+
     def render(self, w, h, st, at):
 
-        if self.start_time is None:
-            self.start_time = renpy.game.interface.frame_time
-
-        if self.anim_time is None:
-            self.anim_time = renpy.game.interface.frame_time
-
-        st = renpy.game.interface.frame_time - self.start_time
-        at = renpy.game.interface.frame_time - self.anim_time
+        st, at = self.adjusted_times()
 
         cr = renpy.display.render.render(self.child, w, h, st, at)
         cw, ch = cr.get_size()
@@ -1652,6 +1731,10 @@ class AdjustTimes(Container):
 
         return rv
 
+    def event(self, ev, x, y, st):
+        st, _ = self.adjusted_times()
+        Container.event(self, ev, x, y, st)
+
     def get_placement(self):
         return self.child.get_placement()
 
