@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -396,7 +396,8 @@ def zoom_render(crend, x, y, w, h, zw, zh, bilinear):
     rv.forward = renpy.display.render.Matrix2D(w / zw, 0, 0, h / zh)
     rv.reverse = renpy.display.render.Matrix2D(zw / w, 0, 0, zh / h)
 
-    rv.clipping = True
+    rv.xclipping = True
+    rv.yclipping = True
 
     rv.blit(crend, rv.reverse.transform(-x, -y))
 
