@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -129,6 +129,7 @@ def get_profile(name):
 
 # Cache ########################################################################
 
+
 # A map from screen name to a list of ScreenCache objects. We ensure the cache
 # does not exceed config.screen_cache_size for each screen.
 predict_cache = collections.defaultdict(list)
@@ -156,6 +157,7 @@ class ScreenCache(object):
         if len(pc) > renpy.config.screen_cache_size:
             pc.pop(0)
 
+
 cache_put = ScreenCache
 
 
@@ -433,12 +435,21 @@ class ScreenDisplayable(renpy.display.layout.Container):
 
         return rv
 
+    def _handles_event(self, event):
+        if self.child is None:
+            return
+
+        return self.child._handles_event(event)
+
     def _hide(self, st, at, kind):
 
         if self.phase == HIDE:
             hid = self
         else:
 
+            if (self.child is not None) and (not self.child._handles_event(kind)):
+                return None
+
             updated_screens.discard(self)
             self.update()
 
@@ -448,6 +459,9 @@ class ScreenDisplayable(renpy.display.layout.Container):
             if self.child is None:
                 return None
 
+            if not self.child._handles_event(kind):
+                return None
+
             if self.screen.ast is not None:
                 self.screen.ast.copy_on_change(self.cache.get(0, {}))
 
@@ -466,7 +480,7 @@ class ScreenDisplayable(renpy.display.layout.Container):
             return None
 
         renpy.ui.detached()
-        hid.child = renpy.ui.fixed(focus="_screen_" + "_".join(self.screen_name))
+        hid.child = renpy.ui.default_fixed(focus="_screen_" + "_".join(self.screen_name))
         hid.children = [ hid.child ]
         renpy.ui.close()
 
@@ -568,7 +582,7 @@ class ScreenDisplayable(renpy.display.layout.Container):
         try:
 
             renpy.ui.detached()
-            self.child = renpy.ui.fixed(focus="_screen_" + "_".join(self.screen_name))
+            self.child = renpy.ui.default_fixed(focus="_screen_" + "_".join(self.screen_name))
             self.children = [ self.child ]
 
             self.scope["_scope"] = self.scope
@@ -580,6 +594,8 @@ class ScreenDisplayable(renpy.display.layout.Container):
             renpy.ui.close()
 
         finally:
+            del self.scope["_scope"]
+
             renpy.ui.screen = old_ui_screen
             pop_current_screen()
 
@@ -661,6 +677,7 @@ class ScreenDisplayable(renpy.display.layout.Container):
     def get_phase_name(self):
         return phase_name[self.phase]
 
+
 # The name of the screen that is currently being displayed, or
 # None if no screen is being currently displayed.
 _current_screen = None
@@ -678,6 +695,7 @@ def push_current_screen(screen):
 def pop_current_screen():
     _current_screen = current_screen_stack.pop()
 
+
 # A map from (screen_name, variant) tuples to screen.
 screens = { }
 
@@ -723,6 +741,7 @@ def get_all_screen_variants(name):
 
     return rv
 
+
 # Have all screens been analyzed?
 analyzed = False
 
@@ -751,6 +770,8 @@ def sort_screens():
     # For each screen, the set of screens that use it.
     reverse = collections.defaultdict(set)
 
+    names = { i[0] for i in screens }
+
     for k, v in screens.items():
 
         name = k[0]
@@ -762,6 +783,10 @@ def sort_screens():
             continue
 
         def callback(uses):
+
+            if uses not in names:
+                return
+
             depends[name].add(uses)
             reverse[uses].add(name)
 
@@ -920,8 +945,7 @@ def get_screen(name, layer=None):
     This can also take a list of names, in which case the first screen
     that is showing is returned.
 
-    This function can be used to check if a screen is showing, using code
-    like::
+    This function can be used to check if a screen is showing::
 
         if renpy.get_screen("say"):
             text "The say screen is showing."
@@ -944,7 +968,9 @@ def get_screen(name, layer=None):
         if sd is not None:
             return sd
 
-        sd = sl.get_displayable_by_name(layer, name)
+    for tag in name:
+
+        sd = sl.get_displayable_by_name(layer, (tag, ))
         if sd is not None:
             return sd
 
@@ -1027,7 +1053,7 @@ def show_screen(_screen_name, *_args, **kwargs):
 
     d = ScreenDisplayable(screen, _tag, _layer, _widget_properties, scope)
 
-    old_d = get_screen(name, _layer)
+    old_d = get_screen(_tag, _layer)
 
     if old_d and old_d.cache:
         d.cache = old_d.cache
@@ -1108,6 +1134,9 @@ def predict_screen(_screen_name, *_args, **kwargs):
             print("While predicting screen", _screen_name)
             traceback.print_exc()
 
+    finally:
+        del scope["_scope"]
+
     renpy.ui.reset()
 
 
@@ -1158,7 +1187,10 @@ def use_screen(_screen_name, *_args, **kwargs):
     scope["_scope"] = scope
     scope["_name"] = (_name, name)
 
-    screen.function(**scope)
+    try:
+        screen.function(**scope)
+    finally:
+        del scope["_scope"]
 
     _current_screen.old_transfers = old_transfers
 
@@ -1203,12 +1235,11 @@ def get_widget_properties(id, screen=None, layer=None):  # @ReservedAssignment
 
     Returns the properties for the widget with `id` in the `screen`
     on `layer`. If `screen` is None, returns the properties for the
-    current screen.
-
-    This can be used inside screen code.
+    current screen. This can be used from Python or property code inside
+    a screen.
 
     Note that this returns a dictionary containing the widget properties,
-    and so to get an individual property
+    and so to get an individual property, the dictionary must be accessed.
     """
 
     if screen is None:
@@ -1271,3 +1302,20 @@ def show_overlay_screens(suppress_overlay):
         for i in renpy.config.overlay_screens:
             if get_screen(i) is not None:
                 hide_screen(i)
+
+
+def per_frame():
+    """
+    Called from interact once per frame to invalidate screens we want to
+    update once per frame.
+    """
+
+    for i in renpy.config.per_frame_screens:
+        s = get_screen(i)
+
+        if s is None:
+            continue
+
+        updated_screens.discard(s)
+        renpy.display.render.invalidate(s)
+        s.update()
