@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -31,6 +31,7 @@ from renpy.display.render import blit_lock, IDENTITY, BLIT, DISSOLVE, IMAGEDISSO
 # A map from cached surface to rle version of cached surface.
 rle_cache = weakref.WeakKeyDictionary()
 
+
 class Clipper(object):
     """
     This is used to calculate the clipping rectangle and update rectangles
@@ -171,7 +172,6 @@ class Clipper(object):
         while sized:
             area, x0, y0, x1, y1 = sized.pop()
 
-
             merged = False
 
             if nca + area >= sa:
@@ -232,8 +232,10 @@ class Clipper(object):
 
         return (x0, y0, x1 - x0, y1 - y0), updates
 
+
 clippers = [ Clipper() ]
 
+
 def surface(w, h, alpha):
     """
     Creates a surface that shares a pixel format with the screen. The created
@@ -247,13 +249,15 @@ def surface(w, h, alpha):
 
     return rv.subsurface((2, 2, w, h))
 
+
 def copy_surface(surf):
     w, h = surf.get_size()
     rv = surface(w, h, True)
 
-    renpy.display.accelerator.nogil_copy(surf, rv) # @UndefinedVariable
+    renpy.display.accelerator.nogil_copy(surf, rv)  # @UndefinedVariable
     return rv
 
+
 def draw_special(what, dest, x, y):
     """
     This handles the special drawing operations, such as dissolve and
@@ -428,11 +432,10 @@ def draw(dest, clip, what, xo, yo, screen):
             # what.draw_func(newdest, newx, newy)
             draw_special(what, newdest, newx, newy)
 
-
         return
 
     # Deal with clipping, if necessary.
-    if what.clipping:
+    if what.xclipping or what.yclipping:
 
         if clip:
             cx0, cy0, cx1, cy1 = clip
@@ -490,10 +493,11 @@ def draw(dest, clip, what, xo, yo, screen):
     for child, cxo, cyo, _focus, _main in what.visible_children:
         draw(dest, clip, child, xo + cxo, yo + cyo, screen)
 
+
 def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
 
     # If our alpha has hit 0, don't do anything.
-    if alpha <= 0.003: # (1 / 256)
+    if alpha <= 0.003:  # (1 / 256)
         return
 
     if forward is None:
@@ -564,7 +568,7 @@ def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
     if what.text_input:
         renpy.display.interface.text_rect = what.screen_rect(xo, yo, reverse)
 
-    if what.clipping:
+    if what.xclipping or what.yclipping:
 
         if reverse.xdy or reverse.ydx:
             draw_transformed(dest, clip, what.pygame_surface(True), xo, yo, alpha, forward, reverse)
@@ -638,7 +642,6 @@ def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
         draw_transformed(dest, clip, child, xo + cxo, yo + cyo, alpha * what.alpha * what.over, child_forward, child_reverse)
 
 
-
 def do_draw_screen(screen_render, full_redraw, swdraw):
     """
     Draws the render produced by render_screen to the screen.
@@ -690,7 +693,6 @@ class SWDraw(object):
         self.mouse_backing_pos = None
         self.mouse_info = None
 
-
         # Is the mouse currently visible?
         self.mouse_old_visible = None
 
@@ -722,6 +724,9 @@ class SWDraw(object):
         # Did we show fullscreen video in the last frame?
         self.showing_video = False
 
+    def get_texture_size(self):
+        return 0, 0
+
     def set_mode(self, virtual_size, physical_size, fullscreen):
 
         # Reset before resize.
@@ -858,7 +863,7 @@ class SWDraw(object):
         if self.mouse_location:
             updates.append(self.hide_mouse())
 
-        if tex and pos and renpy.game.interface.mouse_focused: # @UndefinedVariable
+        if tex and pos and renpy.game.interface.mouse_focused:  # @UndefinedVariable
             updates.append(self.show_mouse(pos, info))
 
         return updates
@@ -913,7 +918,10 @@ class SWDraw(object):
 
         return self.window
 
-    def should_redraw(self, needs_redraw, first_pass):
+    def can_block(self):
+        return True
+
+    def should_redraw(self, needs_redraw, first_pass, can_block):
         """
         Uses the framerate to determine if we can and should redraw.
         """
@@ -949,7 +957,6 @@ class SWDraw(object):
 
         return True
 
-
     def draw_screen(self, surftree, fullscreen_video):
         """
         Draws the screen.
@@ -998,11 +1005,9 @@ class SWDraw(object):
             self.draw_mouse(True)
             pygame.display.flip()
 
-
         if fullscreen_video:
             self.full_redraw = True
 
-
     def render_to_texture(self, render, alpha):
 
         rv = surface(render.width, render.height, alpha)
@@ -1033,7 +1038,6 @@ class SWDraw(object):
             if what.forward:
                 cx, cy = what.forward.transform(cx, cy)
 
-
             if isinstance(child, renpy.display.render.Render):
                 if self.is_pixel_opaque(child, x, y):
                     return True
@@ -1051,7 +1055,6 @@ class SWDraw(object):
 
         return False
 
-
     def mutated_surface(self, surf):
         """
         Called to indicate that the given surface has changed.
@@ -1083,6 +1086,9 @@ class SWDraw(object):
 
         return rle_surf
 
+    def ready_one_texture(self):
+        return False
+
     def solid_texture(self, w, h, color):
         """
         Creates a texture filled to the edges with color.
@@ -1097,7 +1103,6 @@ class SWDraw(object):
         self.mutated_surface(surf)
         return surf
 
-
     def free_memory(self):
         """
         Frees up memory.
@@ -1114,7 +1119,7 @@ class SWDraw(object):
 
         return
 
-    def quit(self): #@ReservedAssignment
+    def quit(self):  # @ReservedAssignment
         """
         Shuts down the drawing system.
         """
