@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -39,6 +39,7 @@ class Viewport(renpy.display.layout.Container):
     __version__ = 5
 
     arrowkeys = False
+    pagekeys = False
 
     def after_upgrade(self, version):
         if version < 1:
@@ -87,6 +88,7 @@ class Viewport(renpy.display.layout.Container):
                  yinitial=None,
                  replaces=None,
                  arrowkeys=False,
+                 pagekeys=False,
                  **properties):
 
         super(Viewport, self).__init__(style=style, **properties)
@@ -118,6 +120,7 @@ class Viewport(renpy.display.layout.Container):
         self.mousewheel = mousewheel
         self.draggable = draggable
         self.arrowkeys = arrowkeys
+        self.pagekeys = pagekeys
 
         # Layout participates in the focus system so drags get migrated.
         self.focusable = draggable or arrowkeys
@@ -202,10 +205,20 @@ class Viewport(renpy.display.layout.Container):
         height = max(height, self.style.yminimum)
 
         if self.set_adjustments:
-            self.xadjustment.range = max(cw - width, 0)
-            self.xadjustment.page = width
-            self.yadjustment.range = max(ch - height, 0)
-            self.yadjustment.page = height
+
+            xarange = max(cw - width, 0)
+
+            if (self.xadjustment.range != xarange) or (self.xadjustment.page != width):
+                self.xadjustment.range = xarange
+                self.xadjustment.page = width
+                self.xadjustment.update()
+
+            yarange = max(ch - height, 0)
+
+            if (self.yadjustment.range != yarange) or (self.yadjustment.page != height):
+                self.yadjustment.range = yarange
+                self.yadjustment.page = height
+                self.yadjustment.update()
 
         if self.xoffset is not None:
             if isinstance(self.xoffset, int):
@@ -258,8 +271,8 @@ class Viewport(renpy.display.layout.Container):
 
         rv = rv.subsurface((0, 0, width, height), focus=True)
 
-        if self.arrowkeys:
-            rv.add_focus(self, None, None, None, None, None)
+        if self.draggable or self.arrowkeys:
+            rv.add_focus(self, None, 0, 0, width, height)
 
         return rv
 
@@ -400,6 +413,24 @@ class Viewport(renpy.display.layout.Container):
                 else:
                     raise renpy.display.core.IgnoreEvent()
 
+        if self.pagekeys:
+
+            if renpy.display.behavior.map_event(ev, 'viewport_pageup'):
+
+                rv = self.yadjustment.change(self.yadjustment.value - self.yadjustment.page)
+                if rv is not None:
+                    return rv
+                else:
+                    raise renpy.display.core.IgnoreEvent()
+
+            if renpy.display.behavior.map_event(ev, 'viewport_pagedown'):
+
+                rv = self.yadjustment.change(self.yadjustment.value + self.yadjustment.page)
+                if rv is not None:
+                    return rv
+                else:
+                    raise renpy.display.core.IgnoreEvent()
+
         if inside and self.draggable:
 
             if renpy.display.behavior.map_event(ev, 'viewport_drag_start'):
@@ -455,6 +486,7 @@ class Viewport(renpy.display.layout.Container):
         self.yoffset = offset
         renpy.display.render.redraw(self, 0)
 
+
 # For compatibility with old saves.
 renpy.display.layout.Viewport = Viewport
 
@@ -507,20 +539,27 @@ class VPGrid(Viewport):
                 rows += 1
 
         # Determine the total size.
-        spacing = self.style.spacing
+        xspacing = self.style.xspacing
+        yspacing = self.style.yspacing
+
+        if xspacing is None:
+            xspacing = self.style.spacing
+        if yspacing is None:
+            yspacing = self.style.spacing
+
         rend = renpy.display.render.render(self.children[0], child_width, child_height, st, at)
         cw, ch = rend.get_size()
 
-        tw = (cw + spacing) * cols - spacing
-        th = (ch + spacing) * rows - spacing
+        tw = (cw + xspacing) * cols - xspacing
+        th = (ch + yspacing) * rows - yspacing
 
         if self.style.xfill:
             tw = child_width
-            cw = (tw - (cols - 1) * spacing) / cols
+            cw = (tw - (cols - 1) * xspacing) / cols
 
         if self.style.yfill:
             th = child_height
-            ch = (th - (rows - 1) * spacing) / rows
+            ch = (th - (rows - 1) * yspacing) / rows
 
         cxo, cyo, width, height = self.update_offsets(tw, th, st)
 
@@ -538,8 +577,8 @@ class VPGrid(Viewport):
                 x = index % cols
                 y = index // cols
 
-            x = x * (cw + spacing) + cxo
-            y = y * (ch + spacing) + cyo
+            x = x * (cw + xspacing) + cxo
+            y = y * (ch + yspacing) + cyo
 
             if x + cw < 0:
                 self.offsets.append((x, y))
@@ -564,7 +603,7 @@ class VPGrid(Viewport):
 
         rv = rv.subsurface((0, 0, width, height), focus=True)
 
-        if self.arrowkeys:
-            rv.add_focus(self, None, None, None, None, None)
+        if self.draggable or self.arrowkeys:
+            rv.add_focus(self, None, 0, 0, width, height)
 
         return rv
