@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -51,10 +51,15 @@ def check_infinite_loop():
 
     global il_time
 
-    if time.time() > il_time:
-        il_time = time.time() + 60
+    now = time.time()
+
+    if now > il_time:
+        il_time = now + 60
         raise Exception("Possible infinite loop.")
 
+    if renpy.config.developer and (il_time > now + 60):
+        il_time = now + 60
+
     return
 
 
@@ -65,6 +70,11 @@ def not_infinite_loop(delay):
     Resets the infinite loop detection timer to `delay` seconds.
     """
 
+    # Give more time in non-developer mode, since computers can be crazy slow
+    # and the player can't do much about it.
+    if not renpy.config.developer:
+        delay *= 5
+
     global il_time
     il_time = time.time() + delay
 
@@ -120,7 +130,7 @@ class Context(renpy.object.Object):
     does participates in rollback.
     """
 
-    __version__ = 13
+    __version__ = 15
 
     nosave = [ 'next_node' ]
 
@@ -172,6 +182,12 @@ class Context(renpy.object.Object):
         if version < 13:
             self.line_log = [ ]
 
+        if version < 14:
+            self.movie = { }
+
+        if version < 15:
+            self.abnormal_stack = [ False ] * len(self.return_stack)
+
     def __init__(self, rollback, context=None, clear=False):
         """
         `clear`
@@ -184,6 +200,9 @@ class Context(renpy.object.Object):
         self.call_location_stack = [ ]
         self.return_stack = [ ]
 
+        # The value of abnormal at the time of the call.
+        self.abnormal_stack = [ ]
+
         # Two deeper then the return stack and call location stack.
         # 1 deeper is for the context top-level, 2 deeper is for
         # _args, _kwargs, and _return.
@@ -235,15 +254,20 @@ class Context(renpy.object.Object):
         # executed?
         self.force_checkpoint = False
 
+        # A mapt from a channel to the Movie playing on that channel.
+        self.movie = { }
+
         if context:
             oldsl = context.scene_lists
             self.runtime = context.runtime
 
             vars(self.info).update(vars(context.info))
 
-            for k, v in context.music.iteritems():
+            for k, v in context.music.items():
                 self.music[k] = v.copy()
 
+            self.movie = dict(context.movie)
+
             self.images = renpy.display.image.ShownImageInfo(context.images)
 
         else:
@@ -323,6 +347,18 @@ class Context(renpy.object.Object):
         while self.dynamic_stack:
             self.pop_dynamic()
 
+    def pop_dynamic_roots(self, roots):
+
+        for dynamic in reversed(self.dynamic_stack):
+
+            for k, v in dynamic.iteritems():
+                name = "store." + k
+
+                if isinstance(v, Delete) and (name in roots):
+                    del roots[name]
+                else:
+                    roots[name] = v
+
     def goto_label(self, node_name):
         """
         Sets the name of the node that will be run when this context
@@ -403,6 +439,11 @@ class Context(renpy.object.Object):
 
         while node:
 
+            this_node = node
+            type_node_name = type(node).__name__
+
+            renpy.plog(1, "--- start {} ({}:{})", type_node_name, node.filename, node.linenumber)
+
             self.current = node.name
             self.last_abnormal = self.abnormal
             self.abnormal = False
@@ -414,12 +455,24 @@ class Context(renpy.object.Object):
                 if ll_entry not in self.line_log:
                     self.line_log.append(ll_entry)
 
-            if self.rollback and renpy.game.log:
-                renpy.game.log.begin()
+            if self.force_checkpoint or (node.rollback == "force"):
+                update_rollback = True
+                force_rollback = True
+            elif not renpy.config.all_nodes_rollback and (node.rollback == "never"):
+                update_rollback = False
+                force_rollback = False
+            else:
+                update_rollback = True
+                force_rollback = False
 
-            if self.force_checkpoint:
-                renpy.game.log.checkpoint(hard=False)
-                self.force_checkpoint = False
+            if update_rollback:
+
+                if self.rollback and renpy.game.log:
+                    renpy.game.log.begin(force=force_rollback)
+
+                if self.force_checkpoint:
+                    renpy.game.log.checkpoint(hard=False)
+                    self.force_checkpoint = False
 
             self.seen = False
 
@@ -435,8 +488,13 @@ class Context(renpy.object.Object):
                     renpy.game.exception_info = "While running game code:"
 
                     self.next_node = None
+
+                    renpy.plog(2, "    before execute {} ({}:{})", type_node_name, node.filename, node.linenumber)
+
                     node.execute()
 
+                    renpy.plog(2, "    after execute {} ({}:{})", type_node_name, node.filename, node.linenumber)
+
                     if developer and self.next_node:
                         self.check_stacks()
 
@@ -471,10 +529,14 @@ class Context(renpy.object.Object):
 
             except renpy.game.CallException as e:
 
-                if self.next_node is None:
-                    raise Exception("renpy.call can't be used when the next node is undefined.")
+                if e.from_current:
+                    return_site = node.name
+                else:
+                    if self.next_node is None:
+                        raise Exception("renpy.call can't be used when the next node is undefined.")
+                    return_site = self.next_node.name
 
-                node = self.call(e.label, return_site=self.next_node.name)
+                node = self.call(e.label, return_site=return_site)
                 self.abnormal = True
                 renpy.store._args = e.args
                 renpy.store._kwargs = e.kwargs
@@ -483,8 +545,10 @@ class Context(renpy.object.Object):
                 renpy.game.persistent._seen_ever[self.current] = True  # @UndefinedVariable
                 renpy.game.seen_session[self.current] = True
 
-            if self.rollback and renpy.game.log:
-                renpy.game.log.complete()
+            renpy.plog(2, "    end {} ({}:{})", type_node_name, this_node.filename, this_node.linenumber)
+
+        if self.rollback and renpy.game.log:
+            renpy.game.log.complete()
 
     def mark_seen(self):
         """
@@ -508,6 +572,7 @@ class Context(renpy.object.Object):
 
         self.return_stack.append(return_site)
         self.dynamic_stack.append({ })
+        self.abnormal_stack.append(self.last_abnormal)
         self.current = label
 
         self.make_dynamic([ "_args", "_kwargs" ])
@@ -530,6 +595,7 @@ class Context(renpy.object.Object):
         self.return_stack.pop()
         self.call_location_stack.pop()
         self.pop_dynamic()
+        self.abnormal_stack.pop()
 
     def lookup_return(self, pop=True):
         """
@@ -554,12 +620,14 @@ class Context(renpy.object.Object):
                 self.return_stack.pop()
                 self.call_location_stack.pop()
                 self.pop_dynamic()
+                self.abnormal = self.abnormal_stack.pop()
 
                 continue
 
             if pop:
                 self.return_stack.pop()
                 self.call_location_stack.pop()
+                self.abnormal = self.abnormal_stack.pop()
 
             return node
 
@@ -582,6 +650,10 @@ class Context(renpy.object.Object):
         rv.translate_language = self.translate_language
         rv.translate_identifier = self.translate_identifier
 
+        rv.abnormal = self.abnormal
+        rv.last_abnormal = self.last_abnormal
+        rv.abnormal_stack = list(self.abnormal_stack)
+
         return rv
 
     def predict_call(self, label, return_site):
@@ -756,6 +828,8 @@ def run_context(top):
 
             if label and renpy.game.script.has_label(label):
                 context.call(label)
+                context.make_dynamic([ "_return" ])
+
             label = None
 
             context.run()
