@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -51,7 +51,10 @@ def renpy_pure(fn):
 
     return fn
 
-from renpy.text.extras import ParameterizedText
+
+import pygame_sdl2
+
+from renpy.text.extras import ParameterizedText, filter_text_tags
 from renpy.text.font import register_sfont, register_mudgefont, register_bmfont
 from renpy.text.text import language_tailor
 from renpy.display.behavior import Keymap
@@ -67,7 +70,7 @@ from renpy.display.screen import get_widget_properties
 from renpy.display.focus import focus_coordinates
 from renpy.display.predict import screen as predict_screen
 
-from renpy.display.image import image_exists, image_exists as has_image
+from renpy.display.image import image_exists, image_exists as has_image, list_images
 from renpy.display.image import get_available_image_tags, get_available_image_attributes, check_image_attributes, get_ordered_image_attributes
 
 from renpy.display.im import load_surface, load_image
@@ -105,6 +108,8 @@ from renpy.sl2.slparser import CustomParser as register_sl_statement, register_s
 
 from renpy.ast import eval_who
 
+from renpy.loader import add_python_directory
+
 renpy_pure("ParameterizedText")
 renpy_pure("Keymap")
 renpy_pure("has_screen")
@@ -114,10 +119,12 @@ renpy_pure("partial")
 renpy_pure("unelide_filename")
 renpy_pure("known_languages")
 renpy_pure("check_text_tags")
+renpy_pure("filter_text_tags")
 
 import time
 import sys
 import threading
+import fnmatch
 
 
 def public_api():
@@ -126,7 +133,7 @@ def public_api():
 
     This does nothing, except to make warnings about unused imports go away.
     """
-    ParameterizedText
+    ParameterizedText, filter_text_tags
     register_sfont, register_mudgefont, register_bmfont
     Keymap
     run, run_action, run_unhovered, run_periodic, map_event
@@ -156,7 +163,7 @@ def public_api():
     check_text_tags
     map_event, queue_event, clear_keymap_cache
     const, pure, not_const
-    image_exists, has_image
+    image_exists, has_image, list_images
     get_available_image_tags, get_available_image_attributes, check_image_attributes, get_ordered_image_attributes
     load_image, load_surface
     profile_memory, diff_memory, profile_rollback
@@ -168,6 +175,8 @@ def public_api():
     register_sl_statement, register_sl_displayable
     eval_who
     is_selected, is_sensitive
+    add_python_directory
+
 
 del public_api
 
@@ -477,6 +486,21 @@ def get_showing_tags(layer='master'):
     return renpy.game.context().images.get_showing_tags(layer)
 
 
+def get_attributes(tag, layer=None):
+    """
+    :doc: image_func
+
+    Return a tuple giving the image attributes for the image with `tag`. If
+    the image is now showing, returns None.
+
+    `layer`
+        The layer to check. If None, uses the default layer for `tag`.
+    """
+
+    layer = default_layer(layer, tag)
+    return renpy.game.context().images.get_attributes(layer, tag, None)
+
+
 def predict_show(name, layer=None, what=None, tag=None, at_list=[ ]):
     """
     :undocumented:
@@ -1072,6 +1096,7 @@ class TagQuotingDict(object):
                 raise Exception("During an interpolation, '%s' was not found as a variable." % key)
             return "<" + key + " unbound>"
 
+
 tag_quoting_dict = TagQuotingDict()
 
 
@@ -1107,7 +1132,7 @@ def scry_say(who, scry):
         scry.interacts = True
 
 
-def say(who, what, interact=True):
+def say(who, what, *args, **kwargs):
     """
     :doc: se_say
 
@@ -1125,6 +1150,7 @@ def say(who, what, interact=True):
     `interact`
         If true, Ren'Py waits for player input when displaying the dialogue. If
         false, Ren'Py shows the dialogue, but does not perform an interaction.
+        (This is passed in as a keyword argument.)
 
     This function is rarely necessary, as the following three lines are
     equivalent. ::
@@ -1141,10 +1167,13 @@ def say(who, what, interact=True):
     if who is None:
         who = renpy.store.narrator  # E1101 @UndefinedVariable
 
+    if renpy.config.say_arguments_callback:
+        args, kwargs = renpy.config.say_arguments_callback(who, *args, **kwargs)
+
     if isinstance(who, (str, unicode)):
-        renpy.store.say(who, what, interact=interact)
+        renpy.store.say(who, what, *args, **kwargs)
     else:
-        who(what, interact=interact)
+        who(what, *args, **kwargs)
 
 
 def imagemap(ground, selected, hotspots, unselected=None, overlays=False,
@@ -1259,7 +1288,7 @@ def pause(delay=None, music=None, with_none=None, hard=False, checkpoint=None):
         if newdelay is not None:
             delay = newdelay
 
-    if renpy.game.after_rollback and roll_forward is None:
+    if (delay is not None) and renpy.game.after_rollback and roll_forward is None:
         delay = 0
 
     if delay is None:
@@ -1333,7 +1362,6 @@ def movie_cutscene(filename, delay=None, loops=0, stop_music=True):
         roll_forward = None
 
     rv = renpy.ui.interact(suppress_overlay=True,
-                           show_mouse=False,
                            roll_forward=roll_forward)
 
     # We don't want to put a checkpoint here, as we can't roll back while
@@ -1380,6 +1408,7 @@ def with_statement(trans, always=False, paired=None, clear=True):
 
     return renpy.game.interface.do_with(trans, paired, clear=clear)
 
+
 globals()["with"] = with_statement
 
 
@@ -1399,8 +1428,8 @@ def rollback(force=False, checkpoints=1, defer=False, greedy=True, label=None, a
         will roll back as far as it can, subject to this condition.
 
     `defer`
-        If true, the call will be deferred until code from the main context is
-        executed.
+        If true, the call will be deferred until control returns to the main
+        context.
 
     `greedy`
         If true, rollback will finish just after the previous checkpoint.
@@ -1410,9 +1439,9 @@ def rollback(force=False, checkpoints=1, defer=False, greedy=True, label=None, a
         If not None, a label that is called when rollback completes.
 
     `abnormal`
-        If true, the default, the code executed after the transition is run in
+        If true, the default, script executed after the transition is run in
         an abnormal mode that skips transitions that would have otherwise
-        occured.
+        occured. Abnormal mode ends when an interaction begins.
     """
 
     if defer and len(renpy.game.contexts) > 1:
@@ -1536,14 +1565,20 @@ def reload_script():
     s = get_screen("menu")
 
     if not renpy.store.main_menu:
+
         if s is not None:
             session["_reload_screen"] = s.screen_name[0]
+            session["_reload_screen_args"] = s.scope.get("_args", ())
+            session["_reload_screen_kwargs"] = s.scope.get("_kwargs", { })
 
         renpy.game.call_in_new_context("_save_reload_game")
 
     else:
+
         if s is not None:
             session["_main_menu_screen"] = s.screen_name[0]
+            session["_main_menu_screen_args"] = s.scope.get("_args", ())
+            session["_main_menu_screen_kwargs"] = s.scope.get("_kwargs", { })
 
         utter_restart()
 
@@ -1594,9 +1629,16 @@ def call(label, *args, **kwargs):
     Causes the current Ren'Py statement to terminate, and a jump to a
     `label` to occur. When the jump returns, control will be passed
     to the statement following the current statement.
+
+    `from_current`
+        If true, control will return to the current statement, rather than
+        the statement following the current statement. (This will lead to
+        the current statement being run twice. This must be passed as a
+        keyword argument.)
     """
 
-    raise renpy.game.CallException(label, args, kwargs)
+    from_current = kwargs.pop("from_current", False)
+    raise renpy.game.CallException(label, args, kwargs, from_current=from_current)
 
 
 def return_statement():
@@ -1614,9 +1656,12 @@ def screenshot(filename):
     :doc: other
 
     Saves a screenshot in `filename`.
+
+    Returns True if the screenshot was saved successfully, False if saving
+    failed for some reason.
     """
 
-    renpy.game.interface.save_screenshot(filename)
+    return renpy.game.interface.save_screenshot(filename)
 
 
 @renpy_pure
@@ -1636,12 +1681,24 @@ def version(tuple=False):  # @ReservedAssignment
 
     return renpy.version
 
+
 version_string = renpy.version
 version_only = renpy.version_only
 version_name = renpy.version_name
 version_tuple = renpy.version_tuple
 license = ""  # @ReservedAssignment
 
+try:
+    import platform as _platform
+    platform = "-".join(_platform.platform().split("-")[:2])
+except:
+    if renpy.android:
+        platform = "Android"
+    elif renpy.ios:
+        platform = "iOS"
+    else:
+        platform = "Unknown"
+
 
 def transition(trans, layer=None, always=False, force=False):
     """
@@ -1864,21 +1921,26 @@ def force_full_redraw():
     renpy.game.interface.full_redraw = True
 
 
-def do_reshow_say(who, what, interact=False):
+def do_reshow_say(who, what, interact=False, *args, **kwargs):
 
     if who is not None:
         who = renpy.python.py_eval(who)
 
-    say(who, what, interact=interact)
+    say(who, what, interact=interact, *args, **kwargs)
+
 
 curried_do_reshow_say = curry(do_reshow_say)
 
 
 def get_reshow_say(**kwargs):
+    kw = dict(renpy.store._last_say_kwargs)
+    kw.update(kwargs)
+
     return curried_do_reshow_say(
         renpy.store._last_say_who,
         renpy.store._last_say_what,
-        **kwargs)
+        renpy.store._last_say_args,
+        **kw)
 
 
 def reshow_say(**kwargs):
@@ -1980,6 +2042,15 @@ def file(fn):  # @ReservedAssignment
     return renpy.loader.load(fn)
 
 
+def notl_file(fn):  # @ReservedAssignment
+    """
+    :undocumented:
+
+    Like file, but doesn't search the translation prefix.
+    """
+    return renpy.loader.load(fn, tl=False)
+
+
 @renpy_pure
 def image_size(im):
     """
@@ -2024,18 +2095,25 @@ def get_at_list(name, layer=None):
     return renpy.game.context().scene_lists.at_list[layer].get(tag, None)
 
 
-def show_layer_at(at_list, layer='master'):
+def show_layer_at(at_list, layer='master', reset=True):
     """
     :doc: se_images
+    :name: renpy.show_layer_at
 
     The python equivalent of the ``show layer`` `layer` ``at`` `at_list`
     statement.
+
+    `reset`
+        If true, the transform state is reset to the start when it is shown.
+        If false, the transform state is persisted, allowing the new transform
+        to update that state.
     """
 
     if not isinstance(at_list, list):
         at_list = [ at_list ]
 
-    renpy.game.context().scene_lists.set_layer_at_list(layer, at_list)
+    renpy.game.context().scene_lists.set_layer_at_list(layer, at_list, reset=reset)
+
 
 layer_at_list = show_layer_at
 
@@ -2083,6 +2161,7 @@ def iconify():
 
     renpy.game.interface.iconify()
 
+
 # New context stuff.
 call_in_new_context = renpy.game.call_in_new_context
 curried_call_in_new_context = renpy.curry.curry(renpy.game.call_in_new_context)
@@ -2098,6 +2177,7 @@ renpy_pure("curried_invoke_in_new_context")
 def _error(msg):
     raise Exception(msg)
 
+
 _error_handlers = [ _error ]
 
 
@@ -2176,13 +2256,14 @@ def load_module(name, **kwargs):
     """
     :doc: other
 
-    This loads the Ren'Py module named name. A Ren'Py module consists of Ren'Py code
+    This loads the Ren'Py module named name. A Ren'Py module consists of Ren'Py script
     that is loaded into the usual (store) namespace, contained in a file named
     name.rpym or name.rpymc. If a .rpym file exists, and is newer than the
     corresponding .rpymc file, it is loaded and a new .rpymc file is created.
 
-    All init code in the module is run before this function returns. An error is
-    raised if the module name cannot be found, or is ambiguous.
+    All of the init blocks (and other init-phase code) in the module are run
+    before this function returns. An error is raised if the module name cannot
+    be found, or is ambiguous.
 
     Module loading may only occur from inside an init block.
     """
@@ -2208,7 +2289,10 @@ def load_module(name, **kwargs):
     renpy.store.__dict__.update(kwargs)  # @UndefinedVariable
 
     for prio, node in initcode:  # @UnusedVariable
-        renpy.game.context().run(node)
+        if isinstance(node, renpy.ast.Node):
+            renpy.game.context().run(node)
+        else:
+            node()
 
     context.pop_all_dynamic()
 
@@ -2246,7 +2330,10 @@ def load_string(s, filename="<string>"):
         renpy.game.contexts.append(context)
 
         for prio, node in initcode:  # @UnusedVariable
-            renpy.game.context().run(node)
+            if isinstance(node, renpy.ast.Node):
+                renpy.game.context().run(node)
+            else:
+                node()
 
         context.pop_all_dynamic()
         renpy.game.contexts.pop()
@@ -2275,6 +2362,7 @@ def pop_call():
 
     renpy.game.context().pop_call()
 
+
 pop_return = pop_call
 
 
@@ -2382,6 +2470,7 @@ def get_image_bounds(tag, width=None, height=None, layer='master'):
 
 # User-Defined Displayable stuff.
 
+
 Render = renpy.display.render.Render
 render = renpy.display.render.render
 IgnoreEvent = renpy.display.core.IgnoreEvent
@@ -2438,6 +2527,27 @@ def cache_unpin(*args):
     renpy.store._cache_pin_set = renpy.store._cache_pin_set - new_pins
 
 
+def expand_predict(d):
+    """
+    :undocumented:
+
+    Use the fnmatch function to expland `d` for the purposes of prediction.
+    """
+
+    if not isinstance(d, basestring):
+        return [ d ]
+
+    if not "*" in d:
+        return [ d ]
+
+    if "." in d:
+        l = list_files(False)
+    else:
+        l = list_images()
+
+    return fnmatch.filter(l, d)
+
+
 def start_predict(*args):
     """
     :doc: image_func
@@ -2445,13 +2555,29 @@ def start_predict(*args):
     This function takes one or more displayables as arguments. It causes
     Ren'Py to predict those displayables during every interaction until
     the displayables are removed by :func:`renpy.stop_predict`.
+
+    If a displayable name is a string containing one or more \\*
+    characters, the asterisks are used as a wildcard pattern. If there
+    is at least one . in the string, the pattern is matched against
+    filenames, otherwise it is matched against image names.
+
+    For example::
+
+        $ renpy.start_predict("eileen *")
+
+    starts predicting all images with the name eileen, while::
+
+        $ renpy.start_predict("images/concert*.*")
+
+    matches all files starting with concert in the images directory.
     """
 
     new_predict = renpy.python.RevertableSet(renpy.store._predict_set)
 
-    for d in args:
-        d = renpy.easy.displayable(d)
-        new_predict.add(d)
+    for i in args:
+        for d in expand_predict(i):
+            d = renpy.easy.displayable(d)
+            new_predict.add(d)
 
     renpy.store._predict_set = new_predict
 
@@ -2462,13 +2588,16 @@ def stop_predict(*args):
 
     This function takes one or more displayables as arguments. It causes
     Ren'Py to stop predicting those displayables during every interaction.
+
+    Wildcard patterns can be used as described in :func:`renpy.start_predict`.
     """
 
     new_predict = renpy.python.RevertableSet(renpy.store._predict_set)
 
-    for d in args:
-        d = renpy.easy.displayable(d)
-        new_predict.discard(d)
+    for i in args:
+        for d in expand_predict(i):
+            d = renpy.easy.displayable(d)
+            new_predict.discard(d)
 
     renpy.store._predict_set = new_predict
 
@@ -2564,6 +2693,8 @@ def list_files(common=False):
 
         rv.append(fn)
 
+    rv.sort()
+
     return rv
 
 
@@ -2586,7 +2717,7 @@ def get_renderer_info():
 
     Other, renderer-specific, keys may also exist. The dictionary should
     be treated as immutable. This should only be called once the display
-    has been started (that is, after the init code is finished).
+    has been started (that is, after the init phase has finished).
     """
 
     return renpy.display.draw.info
@@ -3331,3 +3462,50 @@ def get_adjustment(bar_value):
     """
 
     return bar_value.get_adjustment()
+
+
+def get_skipping():
+    """
+    :doc: other
+
+    Returns true if the Ren'Py is skipping, "fast" if Ren'Py is fast skipping,
+    and false if it is not skipping.
+    """
+
+    return renpy.config.skipping
+
+
+def get_texture_size():
+    """
+    :undocumented:
+
+    Returns the number of bytes of memory locked up in OpenGL textures and the
+    number of textures that are defined.
+    """
+
+    return renpy.display.draw.get_texture_size()
+
+
+old_battery = False
+
+
+def get_on_battery():
+    """
+    :other:
+
+    Returns True if Ren'Py is running on a device that is powered by an internal
+    battery, or False if the device is being charged by some external source.
+    """
+
+    global old_battery
+
+    pi = pygame_sdl2.power.get_power_info()  # @UndefinedVariable
+
+    if pi.state == pygame_sdl2.POWERSTATE_UNKNOWN:  # @UndefinedVariable
+        return old_battery
+    elif pi.state == pygame_sdl2.POWERSTATE_ON_BATTERY:  # @UndefinedVariable
+        old_battery = True
+        return True
+    else:
+        old_battery = False
+        return False
