@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -64,6 +64,7 @@ def report(msg, *args):
     print()
     print(out.encode('utf-8'))
 
+
 added = { }
 
 # Reports additional information about a message, the first time it
@@ -222,6 +223,7 @@ def image_exists(name, expression, tag, precise=True):
 
     report("The image named '%s' was not declared.", names)
 
+
 # Only check each file once.
 check_file_cache = { }
 
@@ -511,6 +513,7 @@ def check_style_property_displayable(name, property, d):
             if rest in renpy.style.all_properties:
                 args = d._args.copy(prefix=prefix)
                 dd = d._duplicate(args)
+                dd._unique()
 
                 check_displayable(
                     "{}, property {}".format(name, prefix + property),
@@ -525,7 +528,7 @@ def check_style(name, s):
 
     for p in s.properties:
         for k, v in p.iteritems():
-            
+
             # Treat font specially.
             if k.endswith("font"):
                 if isinstance(v, renpy.text.font.FontGroup):
@@ -797,6 +800,9 @@ characters per block. """.format(
 
         print()
 
+    for i in renpy.config.lint_stats_callbacks:
+        i()
+
     print()
     if renpy.config.developer and (renpy.config.original_developer != "auto"):
         print("Remember to set config.developer to False before releasing.")
