@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -186,6 +186,9 @@ def load(filename):
     loaded.
     """
 
+    if not os.path.exists(filename):
+        return None
+
     # Unserialize the persistent data.
     try:
         f = file(filename, "rb")
@@ -193,6 +196,14 @@ def load(filename):
         f.close()
         persistent = loads(s)
     except:
+        import renpy.display
+
+        try:
+            renpy.display.log.write("Loading persistent.")
+            renpy.display.log.exception()
+        except:
+            pass
+
         return None
 
     persistent._update()
@@ -272,6 +283,7 @@ def dictset_merge(old, new, current):
     current.update(new)
     return current
 
+
 register_persistent("_seen_ever", dictset_merge)
 register_persistent("_seen_images", dictset_merge)
 register_persistent("_seen_audio", dictset_merge)
@@ -322,6 +334,7 @@ def merge(other):
         backup[f] = safe_deepcopy(val)
         persistent._changed[f] = t
 
+
 # The mtime of the most recently processed savefile.
 persistent_mtime = None
 
@@ -377,6 +390,7 @@ def update(force_save=False):
     if need_save:
         save()
 
+
 should_save_persistent = True
 
 
@@ -483,5 +497,6 @@ def MultiPersistent(name):
     rv._filename = fn  # W0201
     return rv
 
+
 renpy.loadsave._MultiPersistent = _MultiPersistent
 renpy.loadsave.MultiPersistent = MultiPersistent
