@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,15 +23,20 @@
 # contained within the script file. It also handles rolling back the
 # game state to some time in the past.
 
+from __future__ import print_function
+
 # Import the python ast module, not ours.
 ast = __import__("ast", { })
 
+# Import the future module itself.
+import __future__
+
 import marshal
 import random
 import weakref
 import re
-import sets
 import sys
+import time
 
 import renpy.audio
 
@@ -41,12 +46,16 @@ import renpy.audio
 # Deleted is a singleton object that's used to represent an object that has
 # been deleted from the store.
 
+
 class StoreDeleted(object):
+
     def __reduce__(self):
         return "deleted"
 
+
 deleted = StoreDeleted()
 
+
 class StoreModule(object):
     """
     This class represents one of the modules containing the store of data.
@@ -68,9 +77,17 @@ class StoreModule(object):
         del self.__dict__[key]
 
 # Used to unpickle a store module.
+
+
 def get_store_module(name):
     return sys.modules[name]
 
+
+from renpy.pydict import DictItems, find_changes
+
+EMPTY_DICT = { }
+
+
 class StoreDict(dict):
     """
     This class represents the dictionary of a store module. It logs
@@ -84,7 +101,7 @@ class StoreDict(dict):
 
         # The value of this dictionary at the start of the current
         # rollback period (when begin() was last called).
-        self.old = { }
+        self.old = DictItems(self)
 
         # The set of variables in this StoreDict that changed since the
         # end of the init phase.
@@ -95,16 +112,16 @@ class StoreDict(dict):
         Called to reset this to its initial conditions.
         """
 
-        self.old = { }
         self.ever_been_changed = set()
         self.clear()
+        self.old = DictItems(self)
 
     def begin(self):
         """
         Called to mark the start of a rollback period.
         """
 
-        self.old = dict(self)
+        self.old = DictItems(self)
 
     def get_changes(self):
         """
@@ -115,18 +132,12 @@ class StoreDict(dict):
         As a side-effect, updates self.ever_been_changed.
         """
 
-        rv = { }
+        new = DictItems(self)
+        rv = find_changes(self.old, new, deleted)
+        self.old = new
 
-        for k in self:
-            if k not in self.old:
-                rv[k] = deleted
-
-        for k, v in self.old.iteritems():
-
-            new_v = self.get(k, deleted)
-
-            if new_v is not v:
-                rv[k] = v
+        if rv is None:
+            return EMPTY_DICT
 
         for k in rv:
             self.ever_been_changed.add(k)
@@ -145,11 +156,17 @@ store_modules = { }
 # run.
 initialized_store_dicts = set()
 
+
 def create_store(name):
     """
     Creates the store with `name`.
     """
 
+    parent, _, var = name.rpartition('.')
+
+    if parent:
+        create_store(parent)
+
     name = str(name)
 
     if name in initialized_store_dicts:
@@ -178,9 +195,9 @@ def create_store(name):
     else:
         store_modules[name] = sys.modules[name] = StoreModule(d)
 
-    # If we're a module in the store, add us to the store.
-    if name.startswith("store."):
-        store_dicts["store"][name[6:]] = sys.modules[name]
+    if parent:
+        store_dicts[parent][var] = sys.modules[name]
+
 
 class StoreBackup():
     """
@@ -199,29 +216,37 @@ class StoreBackup():
         # The contents of ever_been_changed for each store.
         self.ever_been_changed = { }
 
+        for k in store_dicts:
+            self.backup_one(k)
 
-        for k, v in store_dicts.iteritems():
-            self.store[k] = dict(v)
-            self.old[k] = dict(v.old)
-            self.ever_been_changed[k] = set(v.ever_been_changed)
+    def backup_one(self, name):
+
+        d = store_dicts[name]
+
+        self.store[name] = dict(d)
+        self.old[name] = d.old.as_dict()
+        self.ever_been_changed[name] = set(d.ever_been_changed)
+
+    def restore_one(self, name):
+        sd = store_dicts[name]
+
+        sd.clear()
+        sd.update(self.store[name])
+
+        sd.old = DictItems(self.old[name])
+
+        sd.ever_been_changed.clear()
+        sd.ever_been_changed.update(self.ever_been_changed[name])
 
     def restore(self):
 
-        for k, sd in store_dicts.iteritems():
-
-            sd.clear()
-            sd.update(self.store[k])
-
-            sd.old.clear()
-            sd.old.update(self.old[k])
-
-            sd.ever_been_changed.clear()
-            sd.ever_been_changed.update(self.ever_been_changed[k])
-
+        for k in store_dicts:
+            self.restore_one(k)
 
 
 clean_store_backup = None
 
+
 def make_clean_stores():
     """
     Copy the clean stores.
@@ -231,11 +256,12 @@ def make_clean_stores():
 
     for _k, v in store_dicts.iteritems():
 
-        v.old.clear()
         v.ever_been_changed.clear()
+        v.begin()
 
     clean_store_backup = StoreBackup()
 
+
 def clean_stores():
     """
     Revert the store to the clean copy.
@@ -243,8 +269,29 @@ def clean_stores():
 
     clean_store_backup.restore()
 
-##### Code that computes reachable objects, which is used to filter
-##### the rollback list before rollback or serialization.
+
+def clean_store(name):
+    """
+    Reverts the named store to its clean copy.
+    """
+
+    if not name.startswith("store."):
+        name = "store." + name
+
+    clean_store_backup.restore_one(name)
+
+
+def reset_store_changes(name):
+
+    if not name.startswith("store."):
+        name = "store." + name
+
+    sd = store_dicts[name]
+    sd.begin()
+
+# Code that computes reachable objects, which is used to filter
+# the rollback list before rollback or serialization.
+
 
 class NoRollback(object):
     """
@@ -279,7 +326,7 @@ def reached(obj, reachable, wait):
     if idobj in reachable:
         return
 
-    if isinstance(obj, (NoRollback, file)):
+    if isinstance(obj, (NoRollback, real_file)):  # @UndefinedVariable
         reachable[idobj] = 0
         return
 
@@ -316,6 +363,7 @@ def reached(obj, reachable, wait):
 
     # parents.pop()
 
+
 def reached_vars(store, reachable, wait):
     """
     Marks everything reachable from the variables in the store
@@ -337,68 +385,89 @@ def reached_vars(store, reachable, wait):
                 reached(v, reachable, wait)
 
 
-##### Code that replaces literals will calls to magic constructors.
+# Code that replaces literals will calls to magic constructors.
 
 class WrapNode(ast.NodeTransformer):
 
+    def visit_ClassDef(self, n):
+        n = self.generic_visit(n)
+
+        if not n.bases:
+            n.bases.append(ast.Name(id="object", ctx=ast.Load()))
+
+        return n
+
     def visit_SetComp(self, n):
         return ast.Call(
-            func = ast.Name(
+            func=ast.Name(
                 id="__renpy__set__",
                 ctx=ast.Load()
                 ),
-            args = [ self.generic_visit(n) ],
-            keywords = [ ],
-            starargs = None,
-            kwargs = None)
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
+
+    def visit_Set(self, n):
+
+        return ast.Call(
+            func=ast.Name(
+                id="__renpy__set__",
+                ctx=ast.Load()
+                ),
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
 
     def visit_ListComp(self, n):
         return ast.Call(
-            func = ast.Name(
+            func=ast.Name(
                 id="__renpy__list__",
                 ctx=ast.Load()
                 ),
-            args = [ self.generic_visit(n) ],
-            keywords = [ ],
-            starargs = None,
-            kwargs = None)
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
 
     def visit_List(self, n):
         if not isinstance(n.ctx, ast.Load):
             return self.generic_visit(n)
 
         return ast.Call(
-            func = ast.Name(
+            func=ast.Name(
                 id="__renpy__list__",
                 ctx=ast.Load()
                 ),
-            args = [ self.generic_visit(n) ],
-            keywords = [ ],
-            starargs = None,
-            kwargs = None)
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
 
     def visit_DictComp(self, n):
         return ast.Call(
-            func = ast.Name(
+            func=ast.Name(
                 id="__renpy__dict__",
                 ctx=ast.Load()
                 ),
-            args = [ self.generic_visit(n) ],
-            keywords = [ ],
-            starargs = None,
-            kwargs = None)
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
 
     def visit_Dict(self, n):
 
         return ast.Call(
-            func = ast.Name(
+            func=ast.Name(
                 id="__renpy__dict__",
                 ctx=ast.Load()
                 ),
-            args = [ self.generic_visit(n) ],
-            keywords = [ ],
-            starargs = None,
-            kwargs = None)
+            args=[ self.generic_visit(n) ],
+            keywords=[ ],
+            starargs=None,
+            kwargs=None)
+
 
 wrap_node = WrapNode()
 
@@ -419,6 +488,7 @@ def set_filename(filename, offset, tree):
 
 unicode_re = re.compile(ur'[\u0080-\uffff]')
 
+
 def unicode_sub(m):
     """
     If the string s contains a unicode character, make it into a
@@ -444,13 +514,34 @@ def unicode_sub(m):
 
 string_re = re.compile(r'([uU]?[rR]?)("""|"|\'\'\'|\')((\\.|.)*?)\2')
 
+
 def escape_unicode(s):
     if unicode_re.search(s):
         s = string_re.sub(unicode_sub, s)
 
     return s
 
-def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
+
+# Flags used by py_compile.
+old_compile_flags = ( __future__.nested_scopes.compiler_flag
+                      | __future__.with_statement.compiler_flag
+                      )
+
+new_compile_flags = (  old_compile_flags
+                       | __future__.absolute_import.compiler_flag
+                       | __future__.print_function.compiler_flag
+                       | __future__.unicode_literals.compiler_flag
+                       )
+
+
+# A cache for the results of py_compile.
+py_compile_cache = { }
+
+# An old version of the same, that's preserved across reloads.
+old_py_compile_cache = { }
+
+
+def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False, cache=True):
     """
     Compiles the given source code using the supplied codegenerator.
     Lists, List Comprehensions, and Dictionaries are wrapped when
@@ -476,6 +567,9 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
         that would be used.
     """
 
+    if ast_node:
+        cache = False
+
     if isinstance(source, ast.Module):
         return compile(source, filename, mode)
 
@@ -483,6 +577,26 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
         filename = source.filename
         lineno = source.linenumber
 
+    if cache:
+        key = (lineno, filename, unicode(source), mode, renpy.script.MAGIC)
+
+        rv = py_compile_cache.get(key, None)
+        if rv is not None:
+            return rv
+
+        rv = old_py_compile_cache.get(key, None)
+        if rv is not None:
+            old_py_compile_cache[key] = rv
+            return rv
+
+        bytecode = renpy.game.script.bytecode_oldcache.get(key, None)
+        if bytecode is not None:
+
+            renpy.game.script.bytecode_newcache[key] = bytecode
+            rv = marshal.loads(bytecode)
+            py_compile_cache[key] = rv
+            return rv
+
     source = unicode(source)
     source = source.replace("\r", "")
     source = escape_unicode(source)
@@ -490,7 +604,12 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
     try:
         line_offset = lineno - 1
 
-        tree = ast.parse(source, filename, mode)
+        try:
+            flags = new_compile_flags
+            tree = compile(source, filename, mode, ast.PyCF_ONLY_AST | flags, 1)
+        except:
+            flags = old_compile_flags
+            tree = compile(source, filename, mode, ast.PyCF_ONLY_AST | flags, 1)
 
         tree = wrap_node.visit(tree)
 
@@ -502,7 +621,14 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
         if ast_node:
             return tree.body
 
-        return compile(tree, filename, mode)
+        rv = compile(tree, filename, mode, flags, 1)
+
+        if cache:
+            py_compile_cache[key] = rv
+            renpy.game.script.bytecode_newcache[key] = marshal.dumps(rv)
+            renpy.game.script.bytecode_dirty = True
+
+        return rv
 
     except SyntaxError, e:
 
@@ -513,31 +639,31 @@ def py_compile(source, mode, filename='<none>', lineno=1, ast_node=False):
 
 
 def py_compile_exec_bytecode(source, **kwargs):
-    code = py_compile(source, 'exec', **kwargs)
+    code = py_compile(source, 'exec', cache=False, **kwargs)
     return marshal.dumps(code)
 
 
 def py_compile_eval_bytecode(source, **kwargs):
     source = source.strip()
-    code = py_compile(source, 'eval', **kwargs)
+    code = py_compile(source, 'eval', cache=False, **kwargs)
     return marshal.dumps(code)
 
 
-
-##### Classes that are exported in place of the normal list, dict, and
-##### object.
+# Classes that are exported in place of the normal list, dict, and
+# object.
 
 # This is set to True whenever a mutation occurs. The save code uses
 # this to check to see if a background-save is valid.
 mutate_flag = True
 
+
 def mutator(method):
 
     def do_mutation(self, *args, **kwargs):
 
         global mutate_flag
 
-        mutated = renpy.game.log.mutated #@UndefinedVariable
+        mutated = renpy.game.log.mutated  # @UndefinedVariable
 
         if id(self) not in mutated:
             mutated[id(self)] = ( weakref.ref(self), self._clean())
@@ -605,10 +731,10 @@ class CompressedList(object):
             old_end += 1
 
         # Now that we have this, we can put together the object.
-        self.pre = old[0:old_start]
+        self.pre = list.__getslice__(old, 0, old_start)
         self.start = new_start
         self.end = new_end
-        self.post = old[old_end:]
+        self.post = list.__getslice__(old, old_end, len_old)
 
     def decompress(self, new):
         return self.pre + new[self.start:self.end] + self.post
@@ -644,18 +770,23 @@ class RevertableList(list):
     reverse = mutator(list.reverse)
     sort = mutator(list.sort)
 
-    def wrapper(method): # E0213 @NoSelf
+    def wrapper(method):  # E0213 @NoSelf
         def newmethod(*args, **kwargs):
-            return RevertableList(method(*args, **kwargs)) # E1102
+            l = method(*args, **kwargs)
+            return RevertableList(l)
 
         return newmethod
 
     __add__ = wrapper(list.__add__)
     __getslice__ = wrapper(list.__getslice__)
-    __mul__ = wrapper(list.__mul__)
-    __rmul__ = wrapper(list.__rmul__)
 
-    del wrapper
+    def __mul__(self, other):
+        if not isinstance(other, int):
+            raise TypeError("can't multiply sequence by non-int of type '{}'.".format(type(other).__name__))
+
+        return RevertableList(list.__mul__(self, other))
+
+    __rmul__ = __mul__
 
     def _clean(self):
         """
@@ -694,12 +825,15 @@ class RevertableList(list):
         else:
             self[:] = compressed
 
+
 def revertable_range(*args):
     return RevertableList(range(*args))
 
+
 def revertable_sorted(*args, **kwargs):
     return RevertableList(sorted(*args, **kwargs))
 
+
 class RevertableDict(dict):
 
     def __init__(self, *args, **kwargs):
@@ -716,10 +850,11 @@ class RevertableDict(dict):
     pop = mutator(dict.pop)
     popitem = mutator(dict.popitem)
     setdefault = mutator(dict.setdefault)
+    update = mutator(dict.update)
 
-    def list_wrapper(method): # E0213 @NoSelf
+    def list_wrapper(method):  # E0213 @NoSelf
         def newmethod(*args, **kwargs):
-            return RevertableList(method(*args, **kwargs)) # E1102
+            return RevertableList(method(*args, **kwargs))  # E1102
 
         return newmethod
 
@@ -747,7 +882,21 @@ class RevertableDict(dict):
             self[k] = v
 
 
-class RevertableSet(sets.Set):
+class RevertableSet(set):
+
+    def __setstate__(self, state):
+        if isinstance(state, tuple):
+            self.update(state[0].keys())
+        else:
+            self.update(state)
+
+    def __getstate__(self):
+        rv = ({ i : True for i in self}, )
+        return rv
+
+    # Required to ensure that getstate and setstate are called.
+    __reduce__ = object.__reduce__
+    __reduce_ex__ = object.__reduce_ex__
 
     def __init__(self, *args):
         log = renpy.game.log
@@ -755,44 +904,42 @@ class RevertableSet(sets.Set):
         if log is not None:
             log.mutated[id(self)] = None
 
-        sets.Set.__init__(self, *args)
+        set.__init__(self, *args)
 
-    __iand__ = mutator(sets.Set.__iand__)
-    __ior__ = mutator(sets.Set.__ior__)
-    __isub__ = mutator(sets.Set.__isub__)
-    __ixor__ = mutator(sets.Set.__ixor__)
-    add = mutator(sets.Set.add)
-    clear = mutator(sets.Set.clear)
-    difference_update = mutator(sets.Set.difference_update)
-    discard = mutator(sets.Set.discard)
-    intersection_update = mutator(sets.Set.intersection_update)
-    pop = mutator(sets.Set.pop)
-    remove = mutator(sets.Set.remove)
-    symmetric_difference_update = mutator(sets.Set.symmetric_difference_update)
-    union_update = mutator(sets.Set.union_update)
-    update = mutator(sets.Set.update)
+    __iand__ = mutator(set.__iand__)
+    __ior__ = mutator(set.__ior__)
+    __isub__ = mutator(set.__isub__)
+    __ixor__ = mutator(set.__ixor__)
+    add = mutator(set.add)
+    clear = mutator(set.clear)
+    difference_update = mutator(set.difference_update)
+    discard = mutator(set.discard)
+    intersection_update = mutator(set.intersection_update)
+    pop = mutator(set.pop)
+    remove = mutator(set.remove)
+    symmetric_difference_update = mutator(set.symmetric_difference_update)
+    union_update = mutator(set.update)
+    update = mutator(set.update)
 
-    def wrapper(method): # E0213 @NoSelf
+    def wrapper(method):  # @NoSelf
         def newmethod(*args, **kwargs):
-            rv = method(*args, **kwargs) # E1102
-            if isinstance(rv, sets.Set):
+            rv = method(*args, **kwargs)
+            if isinstance(rv, (set, frozenset)):
                 return RevertableSet(rv)
             else:
                 return rv
 
         return newmethod
 
-    __and__ = wrapper(sets.Set.__and__)
-    __copy__ = wrapper(sets.Set.__copy__)
-    __deepcopy__ = wrapper(sets.Set.__deepcopy__)
-    __sub__ = wrapper(sets.Set.__sub__)
-    __xor__ = wrapper(sets.Set.__xor__)
-    __or__ = wrapper(sets.Set.__or__)
-    copy = wrapper(sets.Set.copy)
-    difference = wrapper(sets.Set.difference)
-    intersection = wrapper(sets.Set.intersection)
-    symmetric_difference = wrapper(sets.Set.symmetric_difference)
-    union = wrapper(sets.Set.union)
+    __and__ = wrapper(set.__and__)
+    __sub__ = wrapper(set.__sub__)
+    __xor__ = wrapper(set.__xor__)
+    __or__ = wrapper(set.__or__)
+    copy = wrapper(set.copy)
+    difference = wrapper(set.difference)
+    intersection = wrapper(set.intersection)
+    symmetric_difference = wrapper(set.symmetric_difference)
+    union = wrapper(set.union)
 
     del wrapper
 
@@ -803,8 +950,8 @@ class RevertableSet(sets.Set):
         return clean
 
     def _rollback(self, compressed):
-        sets.Set.clear(self)
-        sets.Set.update(self, compressed)
+        set.clear(self)
+        set.update(self, compressed)
 
 
 class RevertableObject(object):
@@ -818,6 +965,10 @@ class RevertableObject(object):
 
         return self
 
+    def __init__(self, *args, **kwargs):
+        if (args or kwargs) and renpy.config.developer:
+            raise TypeError("object() takes no parameters.")
+
     def __setattr__(self, attr, value):
         object.__setattr__(self, attr, value)
 
@@ -838,7 +989,7 @@ class RevertableObject(object):
         self.__dict__.update(compressed)
 
 
-##### An object that handles deterministic randomness, or something.
+# An object that handles deterministic randomness, or something.
 
 class DetRandom(random.Random):
 
@@ -878,7 +1029,7 @@ class DetRandom(random.Random):
 
         self.stack = [ ]
 
-    def Random(self,seed=None):
+    def Random(self, seed=None):
         """
         Returns a new RNG object separate from the main one.
         """
@@ -887,10 +1038,15 @@ class DetRandom(random.Random):
         new.seed(seed)
         return new
 
+
 rng = DetRandom()
 
-##### This is the code that actually handles the logging and managing
-##### of the rollbacks.
+# This is the code that actually handles the logging and managing
+# of the rollbacks.
+
+generation = time.time()
+serial = 0
+
 
 class Rollback(renpy.object.Object):
     """
@@ -923,11 +1079,14 @@ class Rollback(renpy.object.Object):
 
     __version__ = 4
 
+    identifier = None
+
     def __init__(self):
 
         super(Rollback, self).__init__()
 
         self.context = renpy.game.context().rollback_copy()
+
         self.objects = [ ]
         self.purged = False
         self.random = [ ]
@@ -946,6 +1105,11 @@ class Rollback(renpy.object.Object):
         # decreases.
         self.hard_checkpoint = False
 
+        # A unique identifier for this rollback object.
+
+        global serial
+        self.identifier = (generation, serial)
+        serial += 1
 
     def after_upgrade(self, version):
 
@@ -966,7 +1130,6 @@ class Rollback(renpy.object.Object):
         if version < 4:
             self.hard_checkpoint = self.checkpoint
 
-
     def purge_unreachable(self, reachable, wait):
         """
         Adds objects that are reachable from the store of this
@@ -1007,15 +1170,13 @@ class Rollback(renpy.object.Object):
                 reached(rb, reachable, wait)
             else:
                 if renpy.config.debug:
-                    print "Removing unreachable:", o
-
+                    print("Removing unreachable:", o, file=renpy.log.real_stdout)
                     pass
 
         self.objects = new_objects
 
         return True
 
-
     def rollback(self):
         """
         Reverts the state of the game to what it was at the start of the
@@ -1053,7 +1214,6 @@ class Rollback(renpy.object.Object):
         renpy.game.contexts.append(self.context)
 
 
-
 class RollbackLog(renpy.object.Object):
     """
     This class manages the list of Rollback objects.
@@ -1072,9 +1232,10 @@ class RollbackLog(renpy.object.Object):
     (weakref to object, information needed to rollback that object)
     """
 
-    __version__ = 4
+    __version__ = 5
 
-    nosave = [ 'old_store', 'mutated' ]
+    nosave = [ 'old_store', 'mutated', 'identifier_cache' ]
+    identifier_cache = None
 
     def __init__(self):
 
@@ -1100,6 +1261,10 @@ class RollbackLog(renpy.object.Object):
         # on load.
         self.retain_after_load_flag = False
 
+        # Has there been an interaction since the last time this log was
+        # reset?
+        self.did_interaction = True
+
     def after_setstate(self):
         self.mutated = { }
         self.rolled_forward = False
@@ -1113,23 +1278,45 @@ class RollbackLog(renpy.object.Object):
         if version < 4:
             self.retain_after_load_flag = False
 
-    def begin(self):
+        if version < 5:
+
+            # We changed what the rollback limit represents, so recompute it
+            # here.
+            if self.rollback_limit:
+                nrbl = 0
+
+                for rb in self.log[-self.rollback_limit:]:
+                    if rb.hard_checkpoint:
+                        nrbl += 1
+
+                self.rollback_limit = nrbl
+
+    def begin(self, force=False):
         """
         Called before a node begins executing, to indicate that the
         state needs to be saved for rollbacking.
         """
 
+        self.identifier_cache = None
+
         context = renpy.game.context()
 
         if not context.rollback:
             return
 
-        # If the transient scene list is not empty, then we do
-        # not begin a new rollback, as the TSL will be purged
-        # after a rollback is complete.
-        if not context.scene_lists.transient_is_empty():
+        # We only begin a checkpoint if the previous statement reached a checkpoint,
+        # or an interaction took place. (Or we're forced.)
+        if (not force) and (self.current and not self.current.checkpoint) and (not self.did_interaction):
             return
 
+        self.did_interaction = False
+
+        if self.current is not None:
+            self.complete()
+        else:
+            for sd in store_dicts.itervalues():
+                sd.begin()
+
         # If the log is too long, prune it.
         if len(self.log) > renpy.config.rollback_length:
             self.log = self.log[-renpy.config.rollback_length:]
@@ -1159,10 +1346,6 @@ class RollbackLog(renpy.object.Object):
 
         self.rolled_forward = False
 
-        # Reset the point that changes are relative to.
-        for sd in store_dicts.itervalues():
-            sd.begin()
-
     def complete(self):
         """
         Called after a node is finished executing, before a save
@@ -1206,8 +1389,6 @@ class RollbackLog(renpy.object.Object):
                 # iterating over it.
                 pass
 
-
-
     def get_roots(self):
         """
         Return a map giving the current roots of the store. This is a
@@ -1225,6 +1406,9 @@ class RollbackLog(renpy.object.Object):
                 else:
                     rv[store_name + "." + name] = deleted
 
+        for i in reversed(renpy.game.contexts[1:]):
+            i.pop_dynamic_roots(rv)
+
         return rv
 
     def purge_unreachable(self, roots, wait=None):
@@ -1288,11 +1472,12 @@ class RollbackLog(renpy.object.Object):
         if not renpy.game.context().rollback:
             return
 
-        if self.rollback_limit < renpy.config.hard_rollback_limit:
-            self.rollback_limit += 1
-
         self.current.checkpoint = True
-        self.current.hard_checkpoint = hard
+
+        if hard and (not self.current.hard_checkpoint):
+            if self.rollback_limit < renpy.config.hard_rollback_limit:
+                self.rollback_limit += 1
+            self.current.hard_checkpoint = hard
 
         if self.in_fixed_rollback() and self.forward:
             # use data from the forward stack
@@ -1311,8 +1496,8 @@ class RollbackLog(renpy.object.Object):
                 fwd_name, fwd_data = self.forward[0]
 
                 if (self.current.context.current == fwd_name
-                    and data == fwd_data
-                    and (keep_rollback or self.rolled_forward)
+                        and data == fwd_data
+                        and (keep_rollback or self.rolled_forward)
                     ):
                     self.forward.pop(0)
                 else:
@@ -1357,7 +1542,7 @@ class RollbackLog(renpy.object.Object):
 
         return self.rollback_limit > 0
 
-    def rollback(self, checkpoints, force=False, label=None, greedy=True, on_load=False):
+    def rollback(self, checkpoints, force=False, label=None, greedy=True, on_load=False, abnormal=True):
         """
         This rolls the system back to the first valid rollback point
         after having rolled back past the specified number of checkpoints.
@@ -1381,15 +1566,19 @@ class RollbackLog(renpy.object.Object):
         `on_load`
             Should be true if this rollback is being called in response to a
             load. Used to implement .retain_after_load()
+
+        `abnormal`
+            If true, treats this as an abnormal event, suppresisng rollback
+            and so on.
         """
 
         # If we have exceeded the rollback limit, and don't have force,
         # give up.
-        if checkpoints and not self.rollback_limit > 0 and not force:
+        if checkpoints and (self.rollback_limit <= 0) and (not force):
             return
 
         self.suspend_checkpointing(False)
-            # will always rollback to before suspension
+        # will always rollback to before suspension
 
         self.purge_unreachable(self.get_roots())
 
@@ -1400,7 +1589,7 @@ class RollbackLog(renpy.object.Object):
             rb = self.log.pop()
             revlog.append(rb)
 
-            if rb.checkpoint:
+            if rb.hard_checkpoint:
                 self.rollback_limit -= 1
 
             if rb.hard_checkpoint or (on_load and rb.checkpoint):
@@ -1416,24 +1605,30 @@ class RollbackLog(renpy.object.Object):
 
             # Otherwise, just give up.
 
-            print "Can't find a place to rollback to. Not rolling back."
+            print("Can't find a place to rollback to. Not rolling back.")
 
             revlog.reverse()
             self.log = self.log + revlog
             return
 
+        force_checkpoint = False
+
         # Try to rollback to just after the previous checkpoint.
         while greedy and self.log:
 
             rb = self.log[-1]
 
-            if rb.checkpoint:
-                break
-
             if not renpy.game.script.has_label(rb.context.current):
                 break
 
-            if self.rollback_limit <= 0:
+            if rb.checkpoint:
+
+                # If the last checkpoint is a label, it's likely the start label.
+                # go back to it.
+                if isinstance(rb.context.current, basestring):
+                    force_checkpoint = True
+                    revlog.append(self.log.pop())
+
                 break
 
             revlog.append(self.log.pop())
@@ -1468,7 +1663,7 @@ class RollbackLog(renpy.object.Object):
             self.log.append(retained)
 
         # Disable the next transition, as it's pointless. (Only when not used with a label.)
-        renpy.game.interface.suppress_transition = True
+        renpy.game.interface.suppress_transition = abnormal
 
         # If necessary, reset the RNG.
         if force:
@@ -1476,22 +1671,31 @@ class RollbackLog(renpy.object.Object):
             self.forward = [ ]
 
         # Flag that we're in the transition immediately after a rollback.
-        renpy.game.after_rollback = True
+        renpy.game.after_rollback = abnormal
 
         # Stop the sounds.
         renpy.audio.audio.rollback()
 
-        # Apply defaults.
-        renpy.exports.execute_default_statement()
+        # Stop any hiding in progress.
+        for i in renpy.game.contexts:
+            i.scene_lists.remove_all_hidden()
 
         renpy.game.contexts.extend(other_contexts)
 
         # Restart the context or the top context.
         if replace_context:
-            raise renpy.game.RestartTopContext(label)
-        else:
-            raise renpy.game.RestartContext(label)
 
+            if force_checkpoint:
+                renpy.game.contexts[0].force_checkpoint = True
+
+            raise renpy.game.RestartTopContext(label)
+
+        else:
+
+            if force_checkpoint:
+                renpy.game.context().force_checkpoint = True
+
+            raise renpy.game.RestartContext(label)
 
     def freeze(self, wait=None):
         """
@@ -1530,7 +1734,7 @@ class RollbackLog(renpy.object.Object):
         """
 
         # Fix up old screens.
-        renpy.display.screen.before_restart() # @UndefinedVariable
+        renpy.display.screen.before_restart()  # @UndefinedVariable
 
         # Set us up as the game log.
         renpy.game.log = self
@@ -1558,21 +1762,52 @@ class RollbackLog(renpy.object.Object):
                 store[name] = value
 
         # Now, rollback to an acceptable point.
-        self.rollback(0, force=True, label=label, greedy=False, on_load=True)
+
+        greedy = renpy.session.pop("_greedy_rollback", False)
+        self.rollback(0, force=True, label=label, greedy=greedy, on_load=True)
 
         # Because of the rollback, we never make it this far.
 
+    def build_identifier_cache(self):
 
-def py_exec_bytecode(bytecode, hide=False, globals=None, locals=None, store="store"): #@ReservedAssignment
+        if self.identifier_cache is not None:
+            return
+
+        rollback_limit = self.rollback_limit
+        checkpoints = 1
+
+        self.identifier_cache = { }
+
+        for i in reversed(self.log):
+
+            if i.identifier is not None:
+                if renpy.game.script.has_label(i.context.current):
+                    self.identifier_cache[i.identifier] = checkpoints
+
+            if i.hard_checkpoint:
+                checkpoints += 1
+
+            if i.checkpoint:
+                rollback_limit -= 1
+
+            if not rollback_limit:
+                break
+
+    def get_identifier_checkpoints(self, identifier):
+        self.build_identifier_cache()
+        return self.identifier_cache.get(identifier, None)
+
+
+def py_exec_bytecode(bytecode, hide=False, globals=None, locals=None, store="store"):  # @ReservedAssignment
 
     if hide:
-        locals = { } #@ReservedAssignment
+        locals = { }  # @ReservedAssignment
 
     if globals is None:
-        globals = store_dicts[store] #@ReservedAssignment
+        globals = store_dicts[store]  # @ReservedAssignment
 
     if locals is None:
-        locals = globals #@ReservedAssignment
+        locals = globals  # @ReservedAssignment
 
     exec bytecode in globals, locals
 
@@ -1583,26 +1818,28 @@ def py_exec(source, hide=False, store=None):
         store = store_dicts["store"]
 
     if hide:
-        locals = { } #@ReservedAssignment
+        locals = { }  # @ReservedAssignment
     else:
-        locals = store #@ReservedAssignment
+        locals = store  # @ReservedAssignment
 
     exec py_compile(source, 'exec') in store, locals
 
 
-def py_eval_bytecode(bytecode, globals=None, locals=None): #@ReservedAssignment
+def py_eval_bytecode(bytecode, globals=None, locals=None):  # @ReservedAssignment
 
     if globals is None:
-        globals = store_dicts["store"] #@ReservedAssignment
+        globals = store_dicts["store"]  # @ReservedAssignment
 
     if locals is None:
-        locals = globals #@ReservedAssignment
+        locals = globals  # @ReservedAssignment
 
     return eval(bytecode, globals, locals)
 
-def py_eval(code, globals=None, locals=None): #@ReservedAssignment
+
+def py_eval(code, globals=None, locals=None):  # @ReservedAssignment
     if isinstance(code, basestring):
         code = py_compile(code, 'eval')
+
     return py_eval_bytecode(code, globals, locals)
 
 
@@ -1629,13 +1866,13 @@ def raise_at_location(e, loc):
 class StoreProxy(object):
 
     def __getattr__(self, k):
-        return getattr(renpy.store, k) #@UndefinedVariable
+        return getattr(renpy.store, k)  # @UndefinedVariable
 
     def __setattr__(self, k, v):
-        setattr(renpy.store, k, v) #@UndefinedVariable
+        setattr(renpy.store, k, v)  # @UndefinedVariable
 
     def __delattr__(self, k):
-        delattr(renpy.store, k) #@UndefinedVariable
+        delattr(renpy.store, k)  # @UndefinedVariable
 
 
 # Code for pickling bound methods.
@@ -1649,9 +1886,26 @@ def method_pickle(method):
 
     return method_unpickle, (obj, name)
 
+
 def method_unpickle(obj, name):
     return getattr(obj, name)
 
+# Code for pickling modules.
+
+
+def module_pickle(module):
+    if renpy.config.developer:
+        raise Exception("Could not pickle {!r}.".format(module))
+
+    return module_unpickle, (module.__name__,)
+
+
+def module_unpickle(name):
+    return __import__(name)
+
+
 import copy_reg
 import types
-copy_reg.pickle(types.MethodType, method_pickle, method_unpickle)
+
+copy_reg.pickle(types.MethodType, method_pickle)
+copy_reg.pickle(types.ModuleType, module_pickle)
