@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -31,6 +31,8 @@ filename = None
 new_variable_serial = 0
 
 # Returns the name of a new variable.
+
+
 @contextlib.contextmanager
 def new_variable():
     global new_variable_serial
@@ -39,11 +41,13 @@ def new_variable():
     yield "_%d" % new_variable_serial
     new_variable_serial -= 1
 
+
 def increment_lineno(node, amount):
     for node in ast.walk(node):
         if hasattr(node, 'lineno'):
             node.lineno += amount
 
+
 class LineNumberNormalizer(ast.NodeVisitor):
 
     def __init__(self):
@@ -64,6 +68,7 @@ class LineNumberNormalizer(ast.NodeVisitor):
 # The parser that things are being added to.
 parser = None
 
+
 class Positional(object):
     """
     This represents a positional parameter to a function.
@@ -78,6 +83,7 @@ class Positional(object):
 # Used to generate the documentation
 all_keyword_names = set()
 
+
 class Keyword(object):
     """
     This represents an optional keyword parameter to a function.
@@ -104,6 +110,7 @@ STYLE_PREFIXES = [
     'selected_activate_',
 ]
 
+
 class Style(object):
     """
     This represents a style parameter to a function.
@@ -222,7 +229,6 @@ class Parser(object):
                 rv.extend(c)
                 count += 1
 
-
         return rv
 
     def parse_eval(self, expr, lineno=1):
@@ -236,7 +242,7 @@ class Parser(object):
 
         try:
             rv = ast.parse(expr, 'eval').body[0].value
-        except SyntaxError, e:
+        except SyntaxError as e:
             raise renpy.parser.ParseError(
                 filename,
                 lineno + e[1][1] - 1,
@@ -259,7 +265,7 @@ class Parser(object):
 
         try:
             rv = ast.parse(code, 'exec')
-        except SyntaxError, e:
+        except SyntaxError as e:
 
             raise renpy.parser.ParseError(
                 filename,
@@ -301,6 +307,7 @@ class Parser(object):
 # A singleton value.
 many = renpy.object.Sentinel("many")
 
+
 class FunctionStatementParser(Parser):
     """
     This is responsible for parsing function statements.
@@ -494,139 +501,139 @@ all_statements = [ ]
 childbearing_statements = [ ]
 
 position_property_names = [
-        "anchor",
-        "xanchor",
-        "yanchor",
-        "pos",
-        "xpos",
-        "ypos",
-        "align",
-        "xalign",
-        "yalign",
-        "xoffset",
-        "yoffset",
-        "maximum",
-        "xmaximum",
-        "ymaximum",
-        "area",
-        "clipping",
-        "xfill",
-        "yfill",
-        # no center, since it can conflict with the center transform.
-        "xcenter",
-        "ycenter",
-        "xsize",
-        "ysize",
-        "xysize",
-        "alt",
-        "debug",
-        ]
+    "anchor",
+    "xanchor",
+    "yanchor",
+    "pos",
+    "xpos",
+    "ypos",
+    "align",
+    "xalign",
+    "yalign",
+    "xoffset",
+    "yoffset",
+    "maximum",
+    "xmaximum",
+    "ymaximum",
+    "area",
+    "clipping",
+    "xfill",
+    "yfill",
+    # no center, since it can conflict with the center transform.
+    "xcenter",
+    "ycenter",
+    "xsize",
+    "ysize",
+    "xysize",
+    "alt",
+    "debug",
+    ]
 
 position_properties = [ Style(i) for i in position_property_names ]
 text_position_properties = [ PrefixStyle("text_", i) for i in position_property_names ]
 side_position_properties = [ PrefixStyle("side_", i) for i in position_property_names ]
 
 text_property_names = [
-        "antialias",
-        "vertical",
-        "black_color",
-        "bold",
-        "color",
-        "drop_shadow",
-        "drop_shadow_color",
-        "first_indent",
-        "font",
-        "size",
-        "hyperlink_functions",
-        "italic",
-        "justify",
-        "kerning",
-        "language",
-        "layout",
-        "line_leading",
-        "line_spacing",
-        "minwidth",
-        "min_width",
-        "newline_indent",
-        "outlines",
-        "rest_indent",
-        "ruby_style",
-        "slow_cps",
-        "slow_cps_multiplier",
-        "slow_abortable",
-        "strikethrough",
-        "text_align",
-        "text_y_fudge",
-        "underline",
-        "minimum",
-        "xminimum",
-        "yminimum",
-        ]
+    "antialias",
+    "vertical",
+    "black_color",
+    "bold",
+    "color",
+    "drop_shadow",
+    "drop_shadow_color",
+    "first_indent",
+    "font",
+    "size",
+    "hyperlink_functions",
+    "italic",
+    "justify",
+    "kerning",
+    "language",
+    "layout",
+    "line_leading",
+    "line_spacing",
+    "minwidth",
+    "min_width",
+    "newline_indent",
+    "outlines",
+    "rest_indent",
+    "ruby_style",
+    "slow_cps",
+    "slow_cps_multiplier",
+    "slow_abortable",
+    "strikethrough",
+    "text_align",
+    "text_y_fudge",
+    "underline",
+    "minimum",
+    "xminimum",
+    "yminimum",
+    ]
 
 text_properties = [ Style(i) for i in text_property_names ]
 text_text_properties = [ PrefixStyle("text_", i) for i in text_property_names ]
 
 window_properties = [ Style(i) for i in [
-        "background",
-        "foreground",
-        "left_margin",
-        "right_margin",
-        "bottom_margin",
-        "top_margin",
-        "xmargin",
-        "ymargin",
-        "left_padding",
-        "right_padding",
-        "top_padding",
-        "bottom_padding",
-        "xpadding",
-        "ypadding",
-        "size_group",
-        "minimum",
-        "xminimum",
-        "yminimum",
-        ] ]
+    "background",
+    "foreground",
+    "left_margin",
+    "right_margin",
+    "bottom_margin",
+    "top_margin",
+    "xmargin",
+    "ymargin",
+    "left_padding",
+    "right_padding",
+    "top_padding",
+    "bottom_padding",
+    "xpadding",
+    "ypadding",
+    "size_group",
+    "minimum",
+    "xminimum",
+    "yminimum",
+    ] ]
 
 button_properties = [ Style(i) for i in [
-        "sound",
-        "mouse",
-        "focus_mask",
-        "child",
-        "keyboard_focus",
-        ] ]
+    "sound",
+    "mouse",
+    "focus_mask",
+    "child",
+    "keyboard_focus",
+    ] ]
 
 bar_properties = [ Style(i) for i in [
-        "bar_vertical",
-        "bar_invert",
-        "bar_resizing",
-        "left_gutter",
-        "right_gutter",
-        "top_gutter",
-        "bottom_gutter",
-        "left_bar",
-        "right_bar",
-        "top_bar",
-        "bottom_bar",
-        "thumb",
-        "thumb_shadow",
-        "thumb_offset",
-        "mouse",
-        "unscrollable",
-        "keyboard_focus",
-        ] ]
+    "bar_vertical",
+    "bar_invert",
+    "bar_resizing",
+    "left_gutter",
+    "right_gutter",
+    "top_gutter",
+    "bottom_gutter",
+    "left_bar",
+    "right_bar",
+    "top_bar",
+    "bottom_bar",
+    "thumb",
+    "thumb_shadow",
+    "thumb_offset",
+    "mouse",
+    "unscrollable",
+    "keyboard_focus",
+    ] ]
 
 box_properties = [ Style(i) for i in [
-        "box_layout",
-        "box_wrap",
-        "box_reverse",
-        "order_reverse",
-        "spacing",
-        "first_spacing",
-        "fit_first",
-        "minimum",
-        "xminimum",
-        "yminimum",
-        ] ]
+    "box_layout",
+    "box_wrap",
+    "box_reverse",
+    "order_reverse",
+    "spacing",
+    "first_spacing",
+    "fit_first",
+    "minimum",
+    "xminimum",
+    "yminimum",
+    ] ]
 
 ui_properties = [
     Keyword("at"),
@@ -735,6 +742,8 @@ Keyword("clicked")
 Keyword("hovered")
 Keyword("unhovered")
 Keyword("alternate")
+Keyword("selected")
+Keyword("sensitive")
 add(ui_properties)
 add(position_properties)
 add(window_properties)
@@ -754,6 +763,8 @@ Keyword("hovered")
 Keyword("unhovered")
 Keyword("alternate")
 Keyword("image_style")
+Keyword("selected")
+Keyword("sensitive")
 add(ui_properties)
 add(position_properties)
 add(window_properties)
@@ -769,6 +780,8 @@ Keyword("alternate")
 Keyword("text_style")
 Keyword("substitute")
 Keyword("scope")
+Keyword("selected")
+Keyword("sensitive")
 add(ui_properties)
 add(position_properties)
 add(window_properties)
@@ -802,6 +815,7 @@ for name in [ "bar", "vbar" ]:
 FunctionStatementParser("viewport", "ui.viewport", 1)
 Keyword("child_size")
 Keyword("mousewheel")
+Keyword("arrowkeys")
 Keyword("draggable")
 Keyword("edgescroll")
 Keyword("xadjustment")
@@ -880,6 +894,7 @@ Keyword("drag_joined")
 Keyword("clicked")
 Keyword("hovered")
 Keyword("unhovered")
+Keyword("mouse_drop")
 Style("child")
 add(ui_properties)
 add(position_properties)
@@ -1094,12 +1109,12 @@ class ForParser(Parser):
             rv = self.parse_exec("%s = 0" % counter_name)
 
             rv.append(ast.For(
-                    target=pattern,
-                    iter=expression,
-                    body=children,
-                    orelse=[],
-                    lineno=lineno,
-                    col_offset=0))
+                target=pattern,
+                iter=expression,
+                body=children,
+                orelse=[],
+                lineno=lineno,
+                col_offset=0))
 
         return rv
 
@@ -1207,10 +1222,10 @@ class ScreenLangScreen(renpy.object.Object):
         self.code = None
 
         # The variant of screen we're defining.
-        self.variant = "None" # expr.
+        self.variant = "None"  # expr.
 
         # Should we predict this screen?
-        self.predict = "None" # expr.
+        self.predict = "None"  # expr.
 
         # The parameters this screen takes.
         self.parameters = None
@@ -1251,7 +1266,6 @@ class ScreenLangScreen(renpy.object.Object):
         renpy.python.py_exec_bytecode(self.code.bytecode, locals=scope)
 
 
-
 class ScreenParser(Parser):
 
     def __init__(self):
@@ -1347,6 +1361,7 @@ class ScreenParser(Parser):
 screen_parser = ScreenParser()
 screen_parser.add(all_statements)
 
+
 def parse_screen(l):
     """
     Parses the screen statement.
