@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -541,6 +541,7 @@ def register_bmfont(name=None, size=None, bold=False, italics=False, underline=F
     bmf = BMFont(filename)
     image_fonts[(name, size, bold, italics)] = bmf
 
+
 # A map from face name to ftfont.FTFace
 face_cache = { }
 
@@ -595,6 +596,7 @@ def load_face(fn):
 
     return rv
 
+
 # Caches of fonts.
 image_fonts = { }
 
@@ -660,13 +662,15 @@ def free_memory():
 
     scaled_image_fonts.clear()
     font_cache.clear()
-    face_cache.clear()
 
 
-def load_image_fonts():
+def load_fonts():
     for i in image_fonts.itervalues():
         i.load()
 
+    for i in renpy.config.preload_fonts:
+        load_face(i)
+
 
 class FontGroup(object):
     """
