@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -210,6 +210,7 @@ class ChildOrFixed(Addable):
         if d is not None:
             raise Exception("Did not expect to close %r." % d)
 
+
 # A stack of things we can add to.
 stack = [ ]
 
@@ -234,6 +235,7 @@ def reset():
     at_stack = [ ]
     imagemap_stack = [ ]
 
+
 renpy.game.post_init.append(reset)
 
 
@@ -276,14 +278,13 @@ def interact(type='misc', roll_forward=None, **kwargs):  # @ReservedAssignment
         renpy.config.skipping = None
 
     if len(stack) != 1:
-        raise Exception("ui.interact called with non-empty widget/layer stack. Did you forget a ui.close() somewhere?")
+        raise Exception("ui.interact called with non-empty widget/layer stack. Did you forget a ui.close() somewhere?\nStack was "+('\n'.join([str(item) for item in stack])))
 
     if at_stack:
         raise Exception("ui.interact called with non-empty at stack.")
 
     renpy.game.context().info._current_interact_type = type
     rv = renpy.game.interface.interact(roll_forward=roll_forward, **kwargs)
-    renpy.game.context().mark_seen()
     renpy.game.context().info._last_interact_type = type
 
     if renpy.exports.in_fixed_rollback() and roll_forward is not None:
@@ -393,6 +394,7 @@ def context_enter(w):
 def context_exit(w):
     close(w)
 
+
 NoStylePrefixGiven = renpy.object.Sentinel("NoStylePrefixGiven")
 
 
@@ -576,6 +578,7 @@ def _add(d, **kwargs):
 
     if d._duplicatable:
         d = d._duplicate(None)
+        d._unique()
 
     rv = d
 
@@ -584,6 +587,7 @@ def _add(d, **kwargs):
 
     return rv
 
+
 add = Wrapper(_add)
 
 
@@ -595,6 +599,7 @@ def _implicit_add(d):
 
     return d
 
+
 implicit_add = Wrapper(_implicit_add)
 
 
@@ -603,9 +608,11 @@ def _image(im, **properties):
 
     if d._duplicatable:
         d = d._duplicate(None)
+        d._unique()
 
     return d
 
+
 image = Wrapper(_image)
 
 null = Wrapper(renpy.display.layout.Null)
@@ -613,6 +620,7 @@ text = Wrapper(renpy.text.text.Text, style="text", replaces=True)
 hbox = Wrapper(renpy.display.layout.MultiBox, layout="horizontal", style="hbox", many=True)
 vbox = Wrapper(renpy.display.layout.MultiBox, layout="vertical", style="vbox", many=True)
 fixed = Wrapper(renpy.display.layout.MultiBox, layout="fixed", style="fixed", many=True)
+default_fixed = Wrapper(renpy.display.layout.MultiBox, layout="fixed", many=True)
 grid = Wrapper(renpy.display.layout.Grid, style="grid", many=True)
 side = Wrapper(renpy.display.layout.Side, style="side", many=True)
 
@@ -620,6 +628,7 @@ side = Wrapper(renpy.display.layout.Side, style="side", many=True)
 def _sizer(maxwidth=None, maxheight=None, **properties):
     return renpy.display.layout.Container(xmaximum=maxwidth, ymaximum=maxheight, **properties)
 
+
 sizer = Wrapper(_sizer, one=True)
 window = Wrapper(renpy.display.layout.Window, style="window", one=True, child=None)
 frame = Wrapper(renpy.display.layout.Window, style="frame", one=True, child=None)
@@ -637,6 +646,7 @@ def _key(key, action=None, activate_sound=None):
 
     return renpy.display.behavior.Keymap(activate_sound=activate_sound, **{ key : action})
 
+
 key = Wrapper(_key)
 
 
@@ -826,6 +836,7 @@ def menu(menuitems,
 
     close()
 
+
 input = Wrapper(renpy.display.behavior.Input, exclude='{}', style="input", replaces=True)  # @ReservedAssignment
 
 
@@ -871,6 +882,7 @@ def imagemap_compat(ground,
 
     close()
 
+
 button = Wrapper(renpy.display.behavior.Button, style='button', one=True)
 
 
@@ -900,10 +912,15 @@ def _imagebutton(idle_image=None,
             return b
 
         if auto is not None:
-            return renpy.config.imagemap_auto_function(auto, name)
+            rv = renpy.config.imagemap_auto_function(auto, name)
+            if rv is not None:
+                return rv
 
         if required:
-            raise Exception("Could not find a %s image for imagemap." % name)
+            if auto:
+                raise Exception("Imagebutton does not have a %s image. (auto=%r)." % (name, auto))
+            else:
+                raise Exception("Imagebutton does not have a %s image." % (name, ))
 
         return None
 
@@ -925,6 +942,7 @@ def _imagebutton(idle_image=None,
         selected_activate_image=selected_activate_image,
         **properties)
 
+
 imagebutton = Wrapper(_imagebutton, style="image_button")
 
 
@@ -952,6 +970,7 @@ def _textbutton(label, clicked=None, style=None, text_style=None, substitute=Tru
     rv._composite_parts = [ text ]
     return rv
 
+
 textbutton = Wrapper(_textbutton)
 
 
@@ -972,6 +991,7 @@ def _label(label, style=None, text_style=None, substitute=True, scope=None, **kw
     rv._composite_parts = [ text ]
     return rv
 
+
 label = Wrapper(_label)
 
 adjustment = renpy.display.behavior.Adjustment
@@ -1017,6 +1037,7 @@ def _bar(*args, **properties):
 
     return renpy.display.behavior.Bar(range, value, width, height, **properties)
 
+
 bar = Wrapper(_bar, vertical=False, replaces=True)
 vbar = Wrapper(_bar, vertical=True, replaces=True)
 slider = Wrapper(_bar, style='slider', replaces=True)
@@ -1037,42 +1058,70 @@ def _autobar_interpolate(range, start, end, time, st, at, **properties):  # @Res
     value = start + t * (end - start)
     return renpy.display.behavior.Bar(range, value, None, None, **properties), redraw
 
+
 autobar_interpolate = renpy.curry.curry(_autobar_interpolate)
 
 
 def _autobar(range, start, end, time, **properties):  # @ReservedAssignment
     return renpy.display.layout.DynamicDisplayable(autobar_interpolate(range, start, end, time, **properties))
 
+
 autobar = Wrapper(_autobar)
 transform = Wrapper(renpy.display.motion.Transform, one=True, style='transform')
 
 _viewport = Wrapper(renpy.display.viewport.Viewport, one=True, replaces=True, style='viewport')
 _vpgrid = Wrapper(renpy.display.viewport.VPGrid, many=True, replaces=True, style='vpgrid')
 
+VIEWPORT_SIZE = 32767
 
-def viewport_common(vpfunc, scrollbars=None, **properties):
+
+def viewport_common(vpfunc, _spacing_to_side, scrollbars=None, **properties):
 
     if scrollbars is None:
         return vpfunc(**properties)
 
-    viewport_properties = { }
-    side_properties = { }
+    (vscrollbar_properties, scrollbar_properties, side_properties, viewport_properties, core_properties) = \
+        renpy.easy.split_properties(properties, "vscrollbar_", "scrollbar_", "side_", "viewport_", "")
 
-    for k, v in properties.iteritems():
-        if k.startswith("side_"):
-            side_properties[k[5:]] = v
-        else:
-            viewport_properties[k] = v
+    if renpy.config.position_viewport_side:
+        from renpy.sl2.slproperties import position_property_names
+
+        for k, v in core_properties.items():
+            if k in position_property_names:
+                side_properties[k] = v
+            elif _spacing_to_side and (k == "spacing"):
+                side_properties[k] = v
+            else:
+                viewport_properties[k] = v
+
+    else:
+        viewport_properties.update(core_properties)
+
+    if renpy.config.prefix_viewport_scrollbar_styles and (scrollbars != "vertical"):
+        scrollbar_properties.setdefault("style", prefixed_style("scrollbar"))
+    else:
+        scrollbar_properties.setdefault("style", "scrollbar")
+
+    if renpy.config.prefix_viewport_scrollbar_styles and (scrollbars != "horizontal"):
+        vscrollbar_properties.setdefault("style", prefixed_style("vscrollbar"))
+    else:
+        vscrollbar_properties.setdefault("style", "vscrollbar")
 
     alt = viewport_properties.get("alt", "viewport")
+    scrollbar_properties.setdefault("alt", alt + " horizontal scrollbar")
+    vscrollbar_properties.setdefault("alt", alt + " vertical scrollbar")
 
     if scrollbars == "vertical":
+
+        if renpy.config.scrollbar_child_size:
+            viewport_properties.setdefault("child_size", (None, VIEWPORT_SIZE))
+
         side("c r", **side_properties)
 
         rv = vpfunc(**viewport_properties)
         addable = stack.pop()
 
-        vscrollbar(adjustment=rv.yadjustment, alt=alt + " vertical scrollbar")
+        vscrollbar(adjustment=rv.yadjustment, **vscrollbar_properties)
         close()
 
         stack.append(addable)
@@ -1080,12 +1129,16 @@ def viewport_common(vpfunc, scrollbars=None, **properties):
         return rv
 
     elif scrollbars == "horizontal":
+
+        if renpy.config.scrollbar_child_size:
+            viewport_properties.setdefault("child_size", (VIEWPORT_SIZE, None))
+
         side("c b", **side_properties)
 
         rv = vpfunc(**viewport_properties)
         addable = stack.pop()
 
-        scrollbar(adjustment=rv.xadjustment, alt=alt + " horizontal scrollbar")
+        scrollbar(adjustment=rv.xadjustment, **scrollbar_properties)
         close()
 
         stack.append(addable)
@@ -1094,13 +1147,16 @@ def viewport_common(vpfunc, scrollbars=None, **properties):
 
     else:
 
+        if renpy.config.scrollbar_child_size:
+            viewport_properties.setdefault("child_size", (VIEWPORT_SIZE, VIEWPORT_SIZE))
+
         side("c r b", **side_properties)
 
         rv = vpfunc(**viewport_properties)
         addable = stack.pop()
 
-        vscrollbar(adjustment=rv.yadjustment, alt=alt + " vertical scrollbar")
-        scrollbar(adjustment=rv.xadjustment, alt=alt +" horizontal scrollbar")
+        vscrollbar(adjustment=rv.yadjustment, **vscrollbar_properties)
+        scrollbar(adjustment=rv.xadjustment, **scrollbar_properties)
         close()
 
         stack.append(addable)
@@ -1109,11 +1165,12 @@ def viewport_common(vpfunc, scrollbars=None, **properties):
 
 
 def viewport(**properties):
-    return viewport_common(_viewport, **properties)
+    return viewport_common(_viewport, True, **properties)
 
 
 def vpgrid(**properties):
-    return viewport_common(_vpgrid, **properties)
+    return viewport_common(_vpgrid, False, **properties)
+
 
 conditional = Wrapper(renpy.display.behavior.Conditional, one=True)
 timer = Wrapper(renpy.display.behavior.Timer, replaces=True)
@@ -1157,22 +1214,23 @@ def _imagemap(ground=None, hover=None, insensitive=None, idle=None, selected_hov
             return variable
 
         if auto:
-            fn = renpy.config.imagemap_auto_function(auto, name)
-            if fn is not None:
-                return fn
+            for i in name:
+                fn = renpy.config.imagemap_auto_function(auto, i)
+                if fn is not None:
+                    return fn
 
         if other is not None:
             return other
 
-        raise Exception("Could not find a %s image for imagemap." % name)
+        raise Exception("Could not find a %s image for imagemap." % name[0])
 
-    ground = pick(ground, "ground", None)
-    idle = pick(idle, "idle", ground)
-    selected_idle = pick(selected_idle, "selected_idle", idle)
-    hover = pick(hover, "hover", ground)
-    selected_hover = pick(selected_hover, "selected_hover", hover)
-    insensitive = pick(insensitive, "insensitive", ground)
-    selected_insensitive = pick(selected_insensitive, "selected_insensitive", hover)
+    ground = pick(ground, ( "ground", "idle" ), idle)
+    idle = pick(idle, ( "idle", ), ground)
+    selected_idle = pick(selected_idle, ( "selected_idle", ), idle)
+    hover = pick(hover, ( "hover", ), ground)
+    selected_hover = pick(selected_hover, ( "selected_hover", ), hover)
+    insensitive = pick(insensitive, ("insensitive", ), ground)
+    selected_insensitive = pick(selected_insensitive, ("selected_insensitive", ), hover)
 
     imagemap_stack.append(
         Imagemap(
@@ -1203,6 +1261,7 @@ def _imagemap(ground=None, hover=None, insensitive=None, idle=None, selected_hov
 
     return rv
 
+
 imagemap = Wrapper(_imagemap, imagemap=True, style='imagemap')
 
 
@@ -1256,6 +1315,7 @@ def _hotspot(spot, style='hotspot', **properties):
         style=style,
         **properties)
 
+
 hotspot_with_child = Wrapper(_hotspot, style="hotspot", one=True)
 
 
@@ -1311,6 +1371,7 @@ def _hotbar(spot, adjustment=None, range=None, value=None, **properties):  # @Re
         ymaximum=h,
         **properties)
 
+
 hotbar = Wrapper(_hotbar, style="hotbar", replaces=True)
 
 
@@ -1321,6 +1382,7 @@ def _returns(v):
 
     return v
 
+
 returns = renpy.curry.curry(_returns)
 
 
@@ -1334,6 +1396,7 @@ def _jumps(label, transition=None):
 
     raise renpy.exports.jump(label)
 
+
 jumps = renpy.curry.curry(_jumps)
 
 
@@ -1341,6 +1404,7 @@ def _jumpsoutofcontext(label):
 
     raise renpy.game.JumpOutException(label)
 
+
 jumpsoutofcontext = renpy.curry.curry(_jumpsoutofcontext)
 
 
@@ -1362,6 +1426,7 @@ def gamemenus(*args):
 ##############################################################################
 # The on statement.
 
+
 on = Wrapper(renpy.display.behavior.OnEvent)
 
 ##############################################################################
@@ -1384,6 +1449,7 @@ def screen_id(id_, d):
 ##############################################################################
 # Postamble
 
+
 # Update the wrappers to have names.
 k, v = None, None
 for k, v in globals().iteritems():
